import math
import warnings
from pathlib import Path
from typing import Optional, Union
from .types import Benchmark, Detector, Matcher
import torch.nn as nn
import cv2
import numpy as np
import torch
import torch.nn.functional as F
from PIL import Image


def get_best_device(verbose=False):
    device = torch.device("cpu")
    if torch.cuda.is_available():
        device = torch.device("cuda")
    elif torch.backends.mps.is_available():
        device = torch.device("mps")
    else:
        device = torch.device("cpu")
    if verbose:
        print(f"Fastest device found is: {device}")
    return device


def recover_pose(E, kpts0, kpts1, K0, K1, mask):
    best_num_inliers = 0
    K0inv = np.linalg.inv(K0[:2, :2])
    K1inv = np.linalg.inv(K1[:2, :2])

    kpts0_n = (K0inv @ (kpts0 - K0[None, :2, 2]).T).T
    kpts1_n = (K1inv @ (kpts1 - K1[None, :2, 2]).T).T

    for _E in np.split(E, len(E) / 3):
        n, R, t, _ = cv2.recoverPose(_E, kpts0_n, kpts1_n, np.eye(3), 1e9, mask=mask)
        if n > best_num_inliers:
            best_num_inliers = n
            ret = (R, t, mask.ravel() > 0)
    return ret


# Code taken from https://github.com/PruneTruong/DenseMatching/blob/40c29a6b5c35e86b9509e65ab0cd12553d998e5f/validation/utils_pose_estimation.py
# --- GEOMETRY ---
def estimate_pose(kpts0, kpts1, K0, K1, norm_thresh, conf=0.99999):
    if len(kpts0) < 5:
        return None
    K0inv = np.linalg.inv(K0[:2, :2])
    K1inv = np.linalg.inv(K1[:2, :2])

    kpts0 = (K0inv @ (kpts0 - K0[None, :2, 2]).T).T
    kpts1 = (K1inv @ (kpts1 - K1[None, :2, 2]).T).T
    E, mask = cv2.findEssentialMat(
        kpts0, kpts1, np.eye(3), threshold=norm_thresh, prob=conf
    )

    ret = None
    if E is not None:
        best_num_inliers = 0

        for _E in np.split(E, len(E) / 3):
            n, R, t, _ = cv2.recoverPose(_E, kpts0, kpts1, np.eye(3), 1e9, mask=mask)
            if n > best_num_inliers:
                best_num_inliers = n
                ret = (R, t, mask.ravel() > 0)
    return ret


def get_grid(B, H, W, device=get_best_device()):
    x1_n = torch.meshgrid(
        *[torch.linspace(-1 + 1 / n, 1 - 1 / n, n, device=device) for n in (B, H, W)],
        indexing="ij",
    )
    x1_n = torch.stack((x1_n[2], x1_n[1]), dim=-1).reshape(B, H * W, 2)
    return x1_n


def fast_inv_2x2(matrix, eps=1e-10):
    return (
        1
        / (torch.linalg.det(matrix)[..., None, None] + eps)
        * torch.stack(
            (
                matrix[..., 1, 1],
                -matrix[..., 0, 1],
                -matrix[..., 1, 0],
                matrix[..., 0, 0],
            ),
            dim=-1,
        ).reshape(*matrix.shape)
    )


def extract_patches_from_inds(x: torch.Tensor, inds: torch.Tensor, patch_size: int):
    B, H, W = x.shape
    B, N = inds.shape
    unfolder = nn.Unfold(kernel_size=patch_size, padding=patch_size // 2, stride=1)
    unfolded_x: torch.Tensor = unfolder(x[:, None])  # B x K_H * K_W x H * W
    patches = torch.gather(
        unfolded_x,
        dim=2,
        index=inds[:, None, :].expand(B, patch_size**2, N),
    )  # B x K_H * K_W x N
    return patches


def extract_patches_from_coords(x: torch.Tensor, coords: torch.Tensor, patch_size: int):
    # NOTE: we could also do this by just adding extra coords and grid_sampling more
    # but this is easy, and the results should be similar
    B, H, W = x.shape
    B, N, two = coords.shape
    unfolder = nn.Unfold(kernel_size=patch_size, padding=patch_size // 2, stride=1)
    unfolded_x: torch.Tensor = unfolder(x[:, None])  # B x K_H * K_W x H * W
    patches = F.grid_sample(
        unfolded_x.reshape(B, patch_size**2, H, W),
        coords[:, None],
        mode="bilinear",
        align_corners=False,
    )[:, 0]  # B x K_H * K_W x N
    return patches


def sample_keypoints(
    keypoint_probs: torch.Tensor,
    num_samples=8192,
    device=get_best_device(),
    use_nms=True,
    nms_size=1,
    sample_topk=True,
    increase_coverage=True,
    remove_borders=False,
    return_probs=False,
    coverage_pow=1 / 2,
    coverage_size=51,
    subpixel=False,
    scoremap=None,  # required for subpixel
    subpixel_temp=0.5,
):
    B, H, W = keypoint_probs.shape
    if increase_coverage:
        weights = (
            -(torch.linspace(-2, 2, steps=coverage_size, device=device) ** 2)
        ).exp()[None, None]
        # 10000 is just some number for maybe numerical stability, who knows. :), result is invariant anyway
        local_density_x = F.conv2d(
            (keypoint_probs[:, None] + 1e-6) * 10000,
            weights[..., None, :],
            padding=(0, coverage_size // 2),
        )
        local_density = F.conv2d(
            local_density_x, weights[..., None], padding=(coverage_size // 2, 0)
        )[:, 0]
        keypoint_probs = keypoint_probs * (local_density + 1e-8) ** (-coverage_pow)
    grid = get_grid(B, H, W, device=device).reshape(B, H * W, 2)
    if use_nms:
        keypoint_probs = keypoint_probs * (
            keypoint_probs
            == F.max_pool2d(keypoint_probs, nms_size, stride=1, padding=nms_size // 2)
        )
    if remove_borders:
        frame = torch.zeros_like(keypoint_probs)
        # we hardcode 4px, could do it nicer, but whatever
        frame[..., 4:-4, 4:-4] = 1
        keypoint_probs = keypoint_probs * frame
    if sample_topk:
        inds = torch.topk(keypoint_probs.reshape(B, H * W), k=num_samples).indices
    else:
        inds = torch.multinomial(
            keypoint_probs.reshape(B, H * W), num_samples=num_samples, replacement=False
        )
    kps = torch.gather(grid, dim=1, index=inds[..., None].expand(B, num_samples, 2))
    if subpixel:
        offsets = get_grid(B, nms_size, nms_size).reshape(
            B, nms_size**2, 2
        )  # B x K_H x K_W x 2
        offsets[..., 0] = offsets[..., 0] * nms_size / W
        offsets[..., 1] = offsets[..., 1] * nms_size / H
        keypoint_patch_scores = extract_patches_from_inds(scoremap, inds, nms_size)
        keypoint_patch_probs = (keypoint_patch_scores / subpixel_temp).softmax(
            dim=1
        )  # B x K_H * K_W x N
        keypoint_offsets = torch.einsum("bkn, bkd ->bnd", keypoint_patch_probs, offsets)
        kps = kps + keypoint_offsets
    if return_probs:
        return kps, torch.gather(keypoint_probs.reshape(B, H * W), dim=1, index=inds)
    return kps


def get_gt_warp(
    depth1,
    depth2,
    T_1to2,
    K1,
    K2,
    depth_interpolation_mode="bilinear",
    relative_depth_error_threshold=0.05,
    H=None,
    W=None,
) -> tuple[torch.Tensor, torch.Tensor]:
    if H is None:
        B, H, W = depth1.shape
    else:
        B = depth1.shape[0]
    with torch.no_grad():
        x1_n = torch.meshgrid(
            *[
                torch.linspace(-1 + 1 / n, 1 - 1 / n, n, device=depth1.device)
                for n in (B, H, W)
            ],
            indexing="ij",
        )
        x1_n = torch.stack((x1_n[2], x1_n[1]), dim=-1).reshape(B, H * W, 2)
        mask, x2 = warp_kpts(
            x1_n.double(),
            depth1.double(),
            depth2.double(),
            T_1to2.double(),
            K1.double(),
            K2.double(),
            depth_interpolation_mode=depth_interpolation_mode,
            relative_depth_error_threshold=relative_depth_error_threshold,
        )
        prob = mask.float().reshape(B, H, W)
        x2 = x2.reshape(B, H, W, 2)
        return torch.cat((x1_n.reshape(B, H, W, 2), x2), dim=-1), prob


def unnormalize_coords(x_n, h, w):
    x = torch.stack(
        (w * (x_n[..., 0] + 1) / 2, h * (x_n[..., 1] + 1) / 2), dim=-1
    )  # [-1+1/h, 1-1/h] -> [0.5, h-0.5]
    return x


def normalize_coords(x, h, w):
    x = torch.stack(
        (2 * (x[..., 0] / w) - 1, 2 * (x[..., 1] / h) - 1), dim=-1
    )  # [-1+1/h, 1-1/h] -> [0.5, h-0.5]
    return x


def rotate_intrinsic(K, n):
    base_rot = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])
    rot = np.linalg.matrix_power(base_rot, n)
    return rot @ K


def rotate_pose_inplane(i_T_w, rot):
    rotation_matrices = [
        np.array(
            [
                [np.cos(r), -np.sin(r), 0.0, 0.0],
                [np.sin(r), np.cos(r), 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            dtype=np.float32,
        )
        for r in [np.deg2rad(d) for d in (0, 270, 180, 90)]
    ]
    return np.dot(rotation_matrices[rot], i_T_w)


def scale_intrinsics(K, scales):
    scales = np.diag([1.0 / scales[0], 1.0 / scales[1], 1.0])
    return np.dot(scales, K)


def angle_error_mat(R1, R2):
    cos = (np.trace(np.dot(R1.T, R2)) - 1) / 2
    cos = np.clip(cos, -1.0, 1.0)  # numercial errors can make it out of bounds
    return np.rad2deg(np.abs(np.arccos(cos)))


def angle_error_vec(v1, v2):
    n = np.linalg.norm(v1) * np.linalg.norm(v2)
    return np.rad2deg(np.arccos(np.clip(np.dot(v1, v2) / n, -1.0, 1.0)))


def compute_pose_error(T_0to1, R, t):
    R_gt = T_0to1[:3, :3]
    t_gt = T_0to1[:3, 3]
    error_t = angle_error_vec(t.squeeze(), t_gt)
    error_t = np.minimum(error_t, 180 - error_t)  # ambiguity of E estimation
    error_R = angle_error_mat(R, R_gt)
    return error_t, error_R


@torch.no_grad()
def warp_kpts(
    kpts0,
    depth0,
    depth1,
    T_0to1,
    K0,
    K1,
    smooth_mask=False,
    return_relative_depth_error=False,
    depth_interpolation_mode="bilinear",
    relative_depth_error_threshold=0.05,
):
    """Warp kpts0 from I0 to I1 with depth, K and Rt
    Also check covisibility and depth consistency.
    Depth is consistent if relative error < 0.2 (hard-coded).
    # https://github.com/zju3dv/LoFTR/blob/94e98b695be18acb43d5d3250f52226a8e36f839/src/loftr/utils/geometry.py adapted from here
    Args:
        kpts0 (torch.Tensor): [N, L, 2] - <x, y>, should be normalized in (-1,1)
        depth0 (torch.Tensor): [N, H, W],
        depth1 (torch.Tensor): [N, H, W],
        T_0to1 (torch.Tensor): [N, 3, 4],
        K0 (torch.Tensor): [N, 3, 3],
        K1 (torch.Tensor): [N, 3, 3],
    Returns:
        calculable_mask (torch.Tensor): [N, L]
        warped_keypoints0 (torch.Tensor): [N, L, 2] <x0_hat, y1_hat>
    """
    (
        n,
        h,
        w,
    ) = depth0.shape
    if depth_interpolation_mode == "combined":
        # Inspired by approach in inloc, try to fill holes from bilinear interpolation by nearest neighbour interpolation
        if smooth_mask:
            raise NotImplementedError("Combined bilinear and NN warp not implemented")
        valid_bilinear, warp_bilinear = warp_kpts(
            kpts0,
            depth0,
            depth1,
            T_0to1,
            K0,
            K1,
            smooth_mask=smooth_mask,
            return_relative_depth_error=return_relative_depth_error,
            depth_interpolation_mode="bilinear",
            relative_depth_error_threshold=relative_depth_error_threshold,
        )
        valid_nearest, warp_nearest = warp_kpts(
            kpts0,
            depth0,
            depth1,
            T_0to1,
            K0,
            K1,
            smooth_mask=smooth_mask,
            return_relative_depth_error=return_relative_depth_error,
            depth_interpolation_mode="nearest-exact",
            relative_depth_error_threshold=relative_depth_error_threshold,
        )
        nearest_valid_bilinear_invalid = (~valid_bilinear).logical_and(valid_nearest)
        warp = warp_bilinear.clone()
        warp[nearest_valid_bilinear_invalid] = warp_nearest[
            nearest_valid_bilinear_invalid
        ]
        valid = valid_bilinear | valid_nearest
        return valid, warp

    kpts0_depth = F.grid_sample(
        depth0[:, None],
        kpts0[:, :, None],
        mode=depth_interpolation_mode,
        align_corners=False,
    )[:, 0, :, 0]
    kpts0 = torch.stack(
        (w * (kpts0[..., 0] + 1) / 2, h * (kpts0[..., 1] + 1) / 2), dim=-1
    )  # [-1+1/h, 1-1/h] -> [0.5, h-0.5]
    # Sample depth, get calculable_mask on depth != 0
    nonzero_mask = kpts0_depth != 0

    # Unproject
    kpts0_h = (
        torch.cat([kpts0, torch.ones_like(kpts0[:, :, [0]])], dim=-1)
        * kpts0_depth[..., None]
    )  # (N, L, 3)
    kpts0_n = K0.inverse() @ kpts0_h.transpose(2, 1)  # (N, 3, L)
    kpts0_cam = kpts0_n

    # Rigid Transform
    w_kpts0_cam = T_0to1[:, :3, :3] @ kpts0_cam + T_0to1[:, :3, [3]]  # (N, 3, L)
    w_kpts0_depth_computed = w_kpts0_cam[:, 2, :]

    # Project
    w_kpts0_h = (K1 @ w_kpts0_cam).transpose(2, 1)  # (N, L, 3)
    w_kpts0 = w_kpts0_h[:, :, :2] / (
        w_kpts0_h[:, :, [2]] + 1e-4
    )  # (N, L, 2), +1e-4 to avoid zero depth

    # Covisible Check
    h, w = depth1.shape[1:3]
    covisible_mask = (
        (w_kpts0[:, :, 0] > 0)
        * (w_kpts0[:, :, 0] < w - 1)
        * (w_kpts0[:, :, 1] > 0)
        * (w_kpts0[:, :, 1] < h - 1)
    )
    w_kpts0 = torch.stack(
        (2 * w_kpts0[..., 0] / w - 1, 2 * w_kpts0[..., 1] / h - 1), dim=-1
    )  # from [0.5,h-0.5] -> [-1+1/h, 1-1/h]
    # w_kpts0[~covisible_mask, :] = -5 # xd

    w_kpts0_depth = F.grid_sample(
        depth1[:, None],
        w_kpts0[:, :, None],
        mode=depth_interpolation_mode,
        align_corners=False,
    )[:, 0, :, 0]

    relative_depth_error = (
        (w_kpts0_depth - w_kpts0_depth_computed) / w_kpts0_depth
    ).abs()
    if not smooth_mask:
        consistent_mask = relative_depth_error < relative_depth_error_threshold
    else:
        consistent_mask = (-relative_depth_error / smooth_mask).exp()
    valid_mask = nonzero_mask * covisible_mask * consistent_mask
    if return_relative_depth_error:
        return relative_depth_error, w_kpts0
    else:
        return valid_mask, w_kpts0


imagenet_mean = torch.tensor([0.485, 0.456, 0.406])
imagenet_std = torch.tensor([0.229, 0.224, 0.225])


def numpy_to_pil(x: np.ndarray):
    """
    Args:
        x: Assumed to be of shape (h,w,c)
    """
    if isinstance(x, torch.Tensor):
        x = x.detach().cpu().numpy()
    if x.max() <= 1.01:
        x *= 255
    x = x.astype(np.uint8)
    return Image.fromarray(x)


def imgnet_unnormalize(x: torch.Tensor) -> torch.Tensor:
    return x * (imagenet_std[:, None, None].to(x.device)) + (
        imagenet_mean[:, None, None].to(x.device)
    )


def imgnet_normalize(x: torch.Tensor) -> torch.Tensor:
    return (x - imagenet_mean[:, None, None].to(x.device)) / (
        imagenet_std[:, None, None].to(x.device)
    )


def tensor_to_pil(x, unnormalize=False, autoscale=False):
    if unnormalize:
        x = imgnet_unnormalize(x)
    if autoscale:
        if x.max() == x.min():
            warnings.warn("x max == x min, cant autoscale")
        else:
            x = (x - x.min()) / (x.max() - x.min())

    x = x.detach()
    if len(x.shape) > 2:
        x = x.permute(1, 2, 0)
    x = x.cpu().numpy()
    x = np.clip(x, 0.0, 1.0)
    return numpy_to_pil(x)


def to_cuda(batch):
    for key, value in batch.items():
        if isinstance(value, torch.Tensor):
            batch[key] = value.cuda()
    return batch


def to_best_device(batch, device=get_best_device()):
    for key, value in batch.items():
        if isinstance(value, torch.Tensor):
            batch[key] = value.to(device)
    return batch


def to_cpu(batch):
    for key, value in batch.items():
        if isinstance(value, torch.Tensor):
            batch[key] = value.cpu()
    return batch


def get_pose(calib):
    w, h = np.array(calib["imsize"])[0]
    return np.array(calib["K"]), np.array(calib["R"]), np.array(calib["T"]).T, h, w


def compute_relative_pose(R1, t1, R2, t2):
    rots = R2 @ (R1.T)
    trans = -rots @ t1 + t2
    return rots, trans


def to_pixel_coords(normalized_coords, h, w) -> torch.Tensor:
    if normalized_coords.shape[-1] != 2:
        raise ValueError(f"Expected shape (..., 2), but got {normalized_coords.shape}")
    pixel_coords = torch.stack(
        (
            w * (normalized_coords[..., 0] + 1) / 2,
            h * (normalized_coords[..., 1] + 1) / 2,
        ),
        axis=-1,
    )
    return pixel_coords


def to_normalized_coords(pixel_coords, h, w) -> torch.Tensor:
    if pixel_coords.shape[-1] != 2:
        raise ValueError(f"Expected shape (..., 2), but got {pixel_coords.shape}")
    normalized_coords = torch.stack(
        (
            2 * (pixel_coords[..., 0]) / w - 1,
            2 * (pixel_coords[..., 1]) / h - 1,
        ),
        axis=-1,
    )
    return normalized_coords


def warp_to_pixel_coords(warp, h1, w1, h2, w2):
    warp1 = warp[..., :2]
    warp1 = torch.stack(
        (
            w1 * (warp1[..., 0] + 1) / 2,
            h1 * (warp1[..., 1] + 1) / 2,
        ),
        axis=-1,
    )
    warp2 = warp[..., 2:]
    warp2 = torch.stack(
        (
            w2 * (warp2[..., 0] + 1) / 2,
            h2 * (warp2[..., 1] + 1) / 2,
        ),
        axis=-1,
    )
    return torch.cat((warp1, warp2), dim=-1)


def to_homogeneous(x):
    ones = torch.ones_like(x[..., -1:])
    return torch.cat((x, ones), dim=-1)


to_hom = to_homogeneous  # alias


def from_homogeneous(xh, eps=1e-12):
    return xh[..., :-1] / (xh[..., -1:] + eps)


from_hom = from_homogeneous  # alias


def homog_transform(Homog, x):
    xh = to_homogeneous(x)
    yh = (Homog @ xh.mT).mT
    y = from_homogeneous(yh)
    return y


def get_homog_warp(Homog, H, W, device=get_best_device()):
    grid = torch.meshgrid(
        torch.linspace(-1 + 1 / H, 1 - 1 / H, H, device=device),
        torch.linspace(-1 + 1 / W, 1 - 1 / W, W, device=device),
        indexing="ij",
    )

    x_A = torch.stack((grid[1], grid[0]), dim=-1)[None]
    x_A_to_B = homog_transform(Homog, x_A)
    mask = ((x_A_to_B > -1) * (x_A_to_B < 1)).prod(dim=-1).float()
    return torch.cat((x_A.expand(*x_A_to_B.shape), x_A_to_B), dim=-1), mask


def dual_log_softmax_matcher(desc_A, desc_B, inv_temperature=1, normalize=False):
    B, N, C = desc_A.shape
    if normalize:
        desc_A = desc_A / desc_A.norm(dim=-1, keepdim=True)
        desc_B = desc_B / desc_B.norm(dim=-1, keepdim=True)
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    else:
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    logP = corr.log_softmax(dim=-2) + corr.log_softmax(dim=-1)
    return logP


def dual_softmax_matcher(desc_A, desc_B, inv_temperature=1, normalize=False):
    if len(desc_A.shape) < 3:
        desc_A, desc_B = desc_A[None], desc_B[None]
    B, N, C = desc_A.shape
    if normalize:
        desc_A = desc_A / desc_A.norm(dim=-1, keepdim=True)
        desc_B = desc_B / desc_B.norm(dim=-1, keepdim=True)
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    else:
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    P = corr.softmax(dim=-2) * corr.softmax(dim=-1)
    return P


def conditional_softmax_matcher(desc_A, desc_B, inv_temperature=1, normalize=False):
    if len(desc_A.shape) < 3:
        desc_A, desc_B = desc_A[None], desc_B[None]
    B, N, C = desc_A.shape
    if normalize:
        desc_A = desc_A / desc_A.norm(dim=-1, keepdim=True)
        desc_B = desc_B / desc_B.norm(dim=-1, keepdim=True)
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    else:
        corr = torch.einsum("b n c, b m c -> b n m", desc_A, desc_B) * inv_temperature
    P_B_cond_A = corr.softmax(dim=-1)
    P_A_cond_B = corr.softmax(dim=-2)

    return P_A_cond_B, P_B_cond_A


def draw_kpts(im, kpts, radius=2, width=1):
    im = np.array(im)
    # Convert keypoints to numpy array
    kpts_np = kpts.cpu().numpy()

    # Create a copy of the image to draw on
    ret = im.copy()

    # Define green color (BGR format in OpenCV)
    green_color = (0, 255, 0)

    # Draw green plus signs for each keypoint
    for x, y in kpts_np:
        # Convert to integer coordinates
        x, y = int(x), int(y)

        # Draw horizontal line of the plus sign
        cv2.line(ret, (x - radius, y), (x + radius, y), green_color, width)
        # Draw vertical line of the plus sign
        cv2.line(ret, (x, y - radius), (x, y + radius), green_color, width)

    return ret


def masked_log_softmax(logits, mask):
    masked_logits = torch.full_like(logits, -torch.inf)
    masked_logits[mask] = logits[mask]
    log_p = masked_logits.log_softmax(dim=-1)
    return log_p


def masked_softmax(logits, mask):
    masked_logits = torch.full_like(logits, -torch.inf)
    masked_logits[mask] = logits[mask]
    log_p = masked_logits.softmax(dim=-1)
    return log_p


def kde(x, std=0.1, half=True, down=None):
    # use a gaussian kernel to estimate density
    if half:
        x = x.half()  # Do it in half precision TODO: remove hardcoding
    if down is not None:
        scores = (-(torch.cdist(x, x[::down]) ** 2) / (2 * std**2)).exp()
    else:
        scores = (-(torch.cdist(x, x) ** 2) / (2 * std**2)).exp()
    density = scores.sum(dim=-1)
    return density


def midpoint_triangulation_unbatched(v1s_local, v2s_local, T1, T2, return_angles=False):
    R1 = T1[:3, :3]  # 3x3 rotation matrix
    R2 = T2[:3, :3]
    t1 = T1[:3, 3]  # 3x1 translation vector
    t2 = T2[:3, 3]

    # Calculate camera centers (single position for each camera)
    C1 = -torch.matmul(R1.T, t1)  # (3,)
    C2 = -torch.matmul(R2.T, t2)  # (3,)

    # # Transform view vectors from local to world coordinates
    # # World vector = R * local_vector

    v1s_world = F.normalize(v1s_local @ R1)  # (N x 3)
    v2s_world = F.normalize(v2s_local @ R2)  # (N x 3)

    # # Vector between camera centers (broadcast to match number of points)
    b = C2 - C1  # (3,)
    num_points = v1s_local.shape[0]
    bs = b.unsqueeze(0).expand(num_points, -1)  # (N x 3)

    # Compute direction vectors between closest points on rays
    cross1 = torch.cross(v1s_world, v2s_world)  # N x 3
    cross2 = torch.cross(bs, v2s_world)  # N x 3

    # Calculate parameters using cross products
    s = torch.sum(cross2 * cross1, dim=1) / torch.sum(cross1 * cross1, dim=1)
    t = torch.sum(torch.cross(bs, v1s_world) * cross1, dim=1) / torch.sum(
        cross1 * cross1, dim=1
    )

    # Find points on each ray in world coordinates
    P1s = C1.unsqueeze(0) + s.unsqueeze(1) * v1s_world  # (N x 3)
    P2s = C2.unsqueeze(0) + t.unsqueeze(1) * v2s_world  # (N x 3)

    # For parallel rays, use camera midpoints
    # midpoint = (C1 + C2) / 2
    # midpoints = midpoint.unsqueeze(0).expand(num_points, -1)
    midpoint = (P1s + P2s) / 2
    if not return_angles:
        return midpoint
    tri_angles = (
        180 / torch.pi * torch.acos((v1s_world * v2s_world).sum(dim=1).clip(0, 1.0))
    )
    return midpoint, tri_angles


def midpoint_triangulation(
    x_A: torch.Tensor,
    x_B: torch.Tensor,
    T_A: torch.Tensor,
    T_B: torch.Tensor,
    return_angles=False,
):
    batch, num_points, three = x_A.shape
    assert three == 3
    # rotation matrix
    R_A = T_A[..., :3, :3]  # (B x 3 x 3)
    R_B = T_B[..., :3, :3]
    # translation vector
    t_A = T_A[..., :3, 3]  # (B x 3)
    t_B = T_B[..., :3, 3]

    # Calculate camera centers (single position for each camera)
    C_A = (R_A.mT @ -t_A[..., None])[..., 0]  # (B x 3 x 3) * (B x 3 x 1) -> (B x 3)
    C_B = (R_B.mT @ -t_B[..., None])[..., 0]  # (B x 3 x 3) * (B x 3 x 1) -> (B x 3)

    # # Transform view vectors from local to world coordinates
    # # World vector = R * local_vector
    ray_A_world = F.normalize(x_A @ R_A, dim=-1)  # (B x N x 3)
    ray_B_world = F.normalize(x_B @ R_B, dim=-1)  # (B x N x 3)

    # # Vector between camera centers (broadcast to match number of points)
    b = C_B - C_A  # (B x 3 x 1)
    bs = b.reshape(batch, 1, three).expand(batch, num_points, three)  # (B x N x 3)

    # Compute direction vectors between closest points on rays
    cross1 = torch.linalg.cross(ray_A_world, ray_B_world)  # B x N x 3
    cross2 = torch.linalg.cross(bs, ray_B_world)  # B x N x 3
    cross3 = torch.linalg.cross(bs, ray_A_world)  # B x N x 3

    # Calculate parameters using cross products
    denom = torch.sum(cross1 * cross1, dim=-1)  # (B x N x 3) -> (B x N)
    s = torch.sum(cross2 * cross1, dim=-1) / denom  # B x N
    t = torch.sum(cross3 * cross1, dim=-1) / denom  # B x N

    # Find points on each ray in world coordinates
    P_A = (
        C_A[:, None] + s[..., None] * ray_A_world
    )  # (B x 1 x 3), (B x N x 1), (B x N x 3) -> (B, N, 3)
    P_B = (
        C_B[:, None] + t[..., None] * ray_B_world
    )  # (B x 1 x 3), (B x N x 1), (B x N x 3) -> (B, N, 3)

    # For parallel rays, use camera midpoints
    midpoint = (P_A + P_B) / 2  # (B x N x 3)
    if not return_angles:
        return midpoint
    tri_angles = (
        180
        / torch.pi
        * torch.acos((ray_A_world * ray_B_world).sum(dim=-1).clip(0, 1.0))
    )  # B x N
    return midpoint, tri_angles


class SkillIssue(NotImplementedError):
    pass


def calibrate(x: torch.Tensor, K: torch.Tensor):
    # x: ..., 2
    # K: ..., 3, 3
    return to_homogeneous(x) @ K.inverse().mT


def project(X: torch.Tensor, T: torch.Tensor, K: torch.Tensor):
    # X: ..., 3
    # T: ..., 4, 4
    # K: ..., 3, 3
    return from_homogeneous(from_homogeneous(to_homogeneous(X) @ T.mT) @ K.mT)


def eye_like(x):
    C, D = x.shape[-2:]
    if C != D:
        raise ValueError(f"Shape not square: {x.shape}")
    e = torch.eye(D).to(x).expand_as(x)
    return e


def triangulate(x_A, x_B, T_A_to_B, K_A, K_B, method="midpoint", return_angles=False):
    if method != "midpoint":
        raise SkillIssue("You should use midpoint instead")
    T_B = T_A_to_B
    T_A = eye_like(T_B)
    x_A_calib = calibrate(x_A, K_A)
    x_B_calib = calibrate(x_B, K_B)
    result = midpoint_triangulation(
        x_A_calib, x_B_calib, T_A, T_B, return_angles=return_angles
    )
    return result


def visualize_keypoints(img_path, vis_path, detector: Detector, num_keypoints: int):
    img_path, vis_path = Path(img_path), Path(vis_path).with_suffix(".png")
    img = Image.open(img_path)
    detections = detector.detect_from_path(
        img_path, num_keypoints=num_keypoints, return_dense_probs=True
    )
    W, H = img.size
    kps = detections["keypoints"]
    kps = detector.to_pixel_coords(kps, H, W)
    (vis_path).parent.mkdir(parents=True, exist_ok=True)
    Image.fromarray(draw_kpts(img, kps[0])).save(vis_path)
    if detections["dense_probs"] is not None:
        tensor_to_pil(detections["dense_probs"].squeeze().cpu(), autoscale=True).save(
            vis_path.as_posix().replace(".png", "_dense_probs.png")
        )


def run_qualitative_examples(
    *, model: Detector, workspace_path: Union[str, Path], test_num_keypoints
):
    import dad

    workspace_path = Path(workspace_path)
    torch.cuda.empty_cache()
    for im_path in [
        "assets/0015_A.jpg",
        "assets/0015_B.jpg",
        "assets/0032_A.jpg",
        "assets/0032_B.jpg",
        "assets/apprentices.jpg",
        "assets/rectangles_and_circles.png",
    ]:
        visualize_keypoints(
            im_path, 
            workspace_path / "vis" / str(dad.GLOBAL_STEP) / im_path,
            model,
            num_keypoints=test_num_keypoints,
        )
    torch.cuda.empty_cache()


def get_experiment_name(experiment_file: str):
    return (
        Path(experiment_file)
        .relative_to(Path("experiments").absolute())
        .with_suffix("")
        .as_posix()
    )


def get_data_iterator(dataset, sample_weights, batch_size, num_steps):
    sampler = torch.utils.data.WeightedRandomSampler(
        sample_weights, num_samples=batch_size * num_steps, replacement=False
    )
    return iter(
        torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            sampler=sampler,
            num_workers=batch_size,
        )
    )


def run_benchmarks(
    benchmarks: list[Benchmark],
    matcher: Matcher,
    detector: Detector,
    *,
    step: int,
    num_keypoints: Optional[list[int] | int] = None,
    sample_every: Optional[int] = 1,
):
    import wandb

    torch.cuda.empty_cache()
    if isinstance(num_keypoints, int):
        num_keypoints = [num_keypoints]

    for bench in benchmarks:
        wandb.log(
            bench(num_keypoints=num_keypoints, sample_every=sample_every)(
                matcher=matcher,
                detector=detector,
            ),
            step=step,
        )
    torch.cuda.empty_cache()


def estimate_pose_essential(
    kps_A: np.ndarray,
    kps_B: np.ndarray,
    w_A: int,
    h_A: int,
    K_A: np.ndarray,
    w_B: int,
    h_B: int,
    K_B: np.ndarray,
    th: float,
) -> tuple[np.ndarray, np.ndarray]:
    import poselib

    camera1 = {
        "model": "PINHOLE",
        "width": w_A,
        "height": h_A,
        "params": K_A[[0, 1, 0, 1], [0, 1, 2, 2]],
    }
    camera2 = {
        "model": "PINHOLE",
        "width": w_B,
        "height": h_B,
        "params": K_B[[0, 1, 0, 1], [0, 1, 2, 2]],
    }

    pose, res = poselib.estimate_relative_pose(
        kps_A,
        kps_B,
        camera1,
        camera2,
        ransac_opt={
            "max_epipolar_error": th,
        },
    )
    return pose.R, pose.t


def poselib_fundamental(x1, x2, opt):
    import poselib

    F, info = poselib.estimate_fundamental(x1, x2, opt, {})
    inl = info["inliers"]
    return F, inl


def estimate_pose_fundamental(
    kps_A: np.ndarray,
    kps_B: np.ndarray,
    w_A: int,
    h_A: int,
    K_A: np.ndarray,
    w_B: int,
    h_B: int,
    K_B: np.ndarray,
    th: float,
) -> tuple[np.ndarray, np.ndarray]:
    if len(kps_A) < 8:
        return np.eye(3), np.zeros(3)
    F, inl = poselib_fundamental(
        kps_A,
        kps_B,
        opt={
            "max_epipolar_error": th,
        },
    )
    E: np.ndarray = K_B.T @ F @ K_A
    kps_calib_A = from_hom(
        calibrate(torch.from_numpy(kps_A).float(), torch.from_numpy(K_A).float())
    ).numpy()
    kps_calib_B = from_hom(
        calibrate(torch.from_numpy(kps_B).float(), torch.from_numpy(K_B).float())
    ).numpy()
    E = E.astype(np.float64)
    _, R, t, good = cv2.recoverPose(E, kps_calib_A, kps_calib_B)
    t = t[:, 0]
    return R, t


def so2(radians):
    return torch.tensor(
        [
            [math.cos(radians), math.sin(radians), 0],
            [-math.sin(radians), math.cos(radians), 0],
            [0, 0, 1.0],
        ]
    )


def rotate_normalized_points(points: torch.Tensor, angle: float):
    # points are between -1, 1, Nx2
    # angle is float [0, 360]
    radians = angle * math.pi / 180
    rot_mat = so2(radians).to(points)
    return points @ rot_mat[:2, :2].T


def compute_detector_correlation(dets1: torch.Tensor, dets2: torch.Tensor, th: float):
    # det1.shape = (K, 2)
    # K = num keypoints
    d = torch.cdist(dets1, dets2, compute_mode="donot_use_mm_for_euclid_dist")
    d12 = d.amin(dim=1)
    d21 = d.amin(dim=0)
    mnn = (d == d12) * (d == d21)
    corr = mnn.float()
    corr[d > th] = 0.0
    return corr.sum(dim=1).mean(), corr.sum(dim=0).mean()


def cross_entropy(log_p_hat: torch.Tensor, p: torch.Tensor):
    return -(log_p_hat * p).sum(dim=-1)


def kl_div(p: torch.Tensor, log_p_hat: torch.Tensor):
    return cross_entropy(log_p_hat, p) - cross_entropy((p + 1e-12).log(), p)


def generalized_mean(r, p1, p2):
    return (1 / 2 * (p1**r + p2**r)) ** (1 / r)


def setup_experiment(experiment_file, root_workspace_path="workspace", disable_wandb=False):
    import wandb

    experiment_name = get_experiment_name(experiment_file)
    wandb.init(
        project="dad",
        mode="online" if not disable_wandb else "disabled",
        name=experiment_name.replace("/", "-"),
    )
    workspace_path = Path(root_workspace_path) / experiment_name
    workspace_path.mkdir(parents=True, exist_ok=True)
    return workspace_path


def check_not_i16(im):
    if im.mode == "I;16":
        raise NotImplementedError("Can't handle 16 bit images")

def wrap_in_sbatch(command, account, time_alloc = "2-23:00:00"):
    sbatch_command = f"""#!/bin/bash
#SBATCH -A {account}
#SBATCH -t {time_alloc}
#SBATCH -o %j.out
#SBATCH --gpus 1
#SBATCH --nodes 1

# Job script commands follow
# Print some GPU info" \
source .venv/bin/activate
{command}
"""
    return sbatch_command


# %BANNER_BEGIN%
# ---------------------------------------------------------------------
# %COPYRIGHT_BEGIN%
#
#  Magic Leap, Inc. ("COMPANY") CONFIDENTIAL
#
#  Unpublished Copyright (c) 2020
#  Magic Leap, Inc., All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property
# of COMPANY. The intellectual and technical concepts contained herein
# are proprietary to COMPANY and may be covered by U.S. and Foreign
# Patents, patents in process, and are protected by trade secret or
# copyright law.  Dissemination of this information or reproduction of
# this material is strictly forbidden unless prior written permission is
# obtained from COMPANY.  Access to the source code contained herein is
# hereby forbidden to anyone except current COMPANY employees, managers
# or contractors who have executed Confidentiality and Non-disclosure
# agreements explicitly covering such access.
#
# The copyright notice above does not evidence any actual or intended
# publication or disclosure  of  this source code, which includes
# information that is confidential and/or proprietary, and is a trade
# secret, of  COMPANY.   ANY REPRODUCTION, MODIFICATION, DISTRIBUTION,
# PUBLIC  PERFORMANCE, OR PUBLIC DISPLAY OF OR THROUGH USE  OF THIS
# SOURCE CODE  WITHOUT THE EXPRESS WRITTEN CONSENT OF COMPANY IS
# STRICTLY PROHIBITED, AND IN VIOLATION OF APPLICABLE LAWS AND
# INTERNATIONAL TREATIES.  THE RECEIPT OR POSSESSION OF  THIS SOURCE
# CODE AND/OR RELATED INFORMATION DOES NOT CONVEY OR IMPLY ANY RIGHTS
# TO REPRODUCE, DISCLOSE OR DISTRIBUTE ITS CONTENTS, OR TO MANUFACTURE,
# USE, OR SELL ANYTHING THAT IT  MAY DESCRIBE, IN WHOLE OR IN PART.
#
# %COPYRIGHT_END%
# ----------------------------------------------------------------------
# %AUTHORS_BEGIN%
#
#  Originating Authors: Paul-Edouard Sarlin
#
# %AUTHORS_END%
# --------------------------------------------------------------------*/
# %BANNER_END%

import torch

from .superpoint import SuperPoint
from .superglue import SuperGlue


class Matching(torch.nn.Module):
    """ Image Matching Frontend (SuperPoint + SuperGlue) """
    def __init__(self, config={}):
        super().__init__()
        self.superpoint = SuperPoint(config.get('superpoint', {}))
        self.superglue = SuperGlue(config.get('superglue', {}))

    def forward(self, data):
        """ Run SuperPoint (optionally) and SuperGlue
        SuperPoint is skipped if ['keypoints0', 'keypoints1'] exist in input
        Args:
          data: dictionary with minimal keys: ['image0', 'image1']
        """
        pred = {}

        # Extract SuperPoint (keypoints, scores, descriptors) if not provided
        if 'keypoints0' not in data:
            pred0 = self.superpoint({'image': data['image0']})
            pred = {**pred, **{k+'0': v for k, v in pred0.items()}}
        if 'keypoints1' not in data:
            pred1 = self.superpoint({'image': data['image1']})
            pred = {**pred, **{k+'1': v for k, v in pred1.items()}}

        # Batch all features
        # We should either have i) one image per batch, or
        # ii) the same number of local features for all images in the batch.
        data = {**data, **pred}

        for k in data:
            if isinstance(data[k], (list, tuple)):
                data[k] = torch.stack(data[k])

        # Perform the matching
        pred = {**pred, **self.superglue(data)}

        return pred



import torch
import numpy as np
import torch.nn.functional as F

from ..matching.nearest_neighbor import NearestNeighborMatcher
from omegaconf import OmegaConf
from .basemodel import BaseExtractor, MethodType
from ..utils import ops
from typing import TypedDict


class XFeatConfig(TypedDict):
    model_name: str
    top_k: int
    detection_threshold: float
    nms_radius: int
    width_confidence: float
    min_corner_score: float


class XFeat_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DETECT_DESCRIBE
    default_conf: XFeatConfig = {
        "model_name": "xfeat",
        "top_k": 2048,
        "detection_threshold": 0.2,
        "nms_radius": 4,
        "width_confidence": 0.5,
        "min_corner_score": 0.0,
    }

    def __init__(self, conf: XFeatConfig = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        device = "cpu"
        self.device = torch.device(device)
        self.model = torch.hub.load(
            "verlab/accelerated_features",
            "XFeat",
            pretrained=True,
            top_k=conf.top_k,
            detection_threshold=conf.detection_threshold,
        )
        self.model.eval()
        self.model.dev = self.device
        self.model.to(device)
        print(f"Loaded XFeat model to {self.device}")
        print(f"Model device: {self.model.dev}")
        self.matcher = NearestNeighborMatcher()

    def detectAndCompute(self, img, return_dict=None):
        img = ops.prepareImage(img).to(self.model.dev)
        response = self.model.detectAndCompute(img, top_k=self.conf.top_k)

        # batch it
        batch_response = {}
        for im_result in response:
            for key in im_result:
                if key not in batch_response:
                    batch_response[key] = []
                batch_response[key].append(im_result[key])

        for key in batch_response:
            batch_response[key] = torch.stack(batch_response[key])
            if len(batch_response[key].shape) == 2:
                batch_response[key] = batch_response[key].unsqueeze(0)

        if return_dict:
            return batch_response

        return batch_response["keypoints"], batch_response["descriptors"]

    def detect(self, img, op=None):
        img = ops.prepareImage(img).to(self.device)
        return self.detectAndCompute(img, return_dict=True)["keypoints"]

    # def match(self, image1, image2):
    #     mkpts0, mkpts1 = self.model.match_xfeat(image1, image2)
    #     return {
    #         'mkpts0': mkpts0,
    #         'mkpts1': mkpts1,
    #     }

    def match_xfeat(self, image1, image2):
        mkpts0, mkpts1 = self.model.match_xfeat(image1, image2)
        return {
            "mkpts0": mkpts0,
            "mkpts1": mkpts1,
        }

    def match_xfeat_star(self, image1, image2):
        mkpts0, mkpts1 = self.model.match_xfeat_star(image1, image2)
        return {
            "mkpts0": mkpts0,
            "mkpts1": mkpts1,
        }

    def compute(self, img, keypoints):
        """Compute descriptors for keypoints in the image.
        img: np.array, image
        keypoints: np.array | torch.tensor, keypoints
        """

        if isinstance(keypoints, np.ndarray):
            keypoints = torch.tensor(keypoints)
        keypoints = keypoints.to(self.device)

        # if unbatched, add batch dimension
        if len(keypoints.shape) == 2:
            keypoints = keypoints.unsqueeze(0)

        img = ops.prepareImage(img, gray=True).to(self.device)

        x, rh1, rw1 = self.model.preprocess_tensor(img)
        keypoints = keypoints / torch.tensor([rw1, rh1]).to(self.device)
        B, _, _H1, _W1 = x.shape

        M1, K1, H1 = self.model.net(x)
        M1 = F.normalize(M1, dim=1)

        # Interpolate descriptors at kpts positions
        feats = self.model.interpolator(M1, keypoints, H=_H1, W=_W1)

        # L2-Normalize
        feats = F.normalize(feats, dim=-1)
        return feats

    def to(self, device):
        self.model.to(device)
        self.model.dev = device
        self.device = device
        return self

    @property
    def has_detector(self):
        return True


if __name__ == "__main__":
    from easy_local_features.utils import io, ops

    method = XFeat_baseline()

    img0 = io.fromPath("test/assets/megadepth0.jpg")

    kpts = method.detect(img0)
    desc = method.compute(img0, kpts)

    kpts2, desc2 = method.detectAndCompute(img0)

    assert torch.allclose(kpts, kpts2)

    # import pdb; pdb.set_trace()
    print(desc)
    print(desc2)
    assert torch.allclose(desc, desc2, atol=1e-5)

    # img1 = io.fromPath("test/assets/megadepth1.jpg")

    # nn_matches = method.match(img0, img1)
    # xfeat_matches = method.match_xfeat(img0, img1)
    # xfeat_star_matches = method.match_xfeat_star(img0, img1)

    # vis.plot_pair(img0, img1)
    # vis.plot_matches(nn_matches['mkpts0'], nn_matches['mkpts1'])
    # vis.add_text("")

    # vis.plot_pair(img0, img1)
    # vis.plot_matches(xfeat_matches['mkpts0'], xfeat_matches['mkpts1'])

    # vis.plot_pair(img0, img1)
    # vis.plot_matches(xfeat_star_matches['mkpts0'], xfeat_star_matches['mkpts1'])

    # vis.show()



from .dedode_detector import load_DaD as load_DaD
from .dedode_detector import load_DaDDark as load_DaDDark
from .dedode_detector import load_DaDLight as load_DaDLight
from .dedode_detector import dedode_detector_S as dedode_detector_S
from .dedode_detector import dedode_detector_B as dedode_detector_B
from .dedode_detector import dedode_detector_L as dedode_detector_L
from .dedode_detector import load_dedode_v2 as load_dedode_v2


lg_detectors = ["ALIKED", "ALIKEDROT", "SIFT", "DISK", "SuperPoint", "ReinforcedFP"]
other_detectors = ["HesAff", "HarrisAff", "REKD"]
dedode_detectors = [
    "DeDoDe-v2",
    "DaD",
    "DaDLight",
    "DaDDark",
]
all_detectors = lg_detectors + dedode_detectors + other_detectors


def load_detector_by_name(detector_name, *, resize=1024, weights_path=None):
    if detector_name == "DaD":
        detector = load_DaD(resize=resize, weights_path=weights_path)
    elif detector_name == "DaDLight":
        detector = load_DaDLight(resize=resize, weights_path=weights_path)
    elif detector_name == "DaDDark":
        detector = load_DaDDark(resize=resize, weights_path=weights_path)
    elif detector_name == "DeDoDe-v2":
        detector = load_dedode_v2()
    else:
        raise ValueError(f"Couldn't find detector with detector name {detector_name}")
    return detector



'''
A super generic match util.
'''


class GenericMatcher():
    def __init__(self, method_name, **kwargs):
        self.matcher = None
        self.matcher_kwargs = kwargs

    def __call__(self, features1, features2):
        return self.matcher(features1, features2, **self.matcher_kwargs)


"""
DINOv3 baseline (descriptor-only), following DINOv2 patterns and official DINOv3 usage:
- Requires Python >= 3.10 (hubconf uses PEP 604 | unions)
- Load via torch.hub (recommended: local repo + explicit weights)
- Use ImageNet normalization for LVD-1689M weights and SAT normalization for SAT-493M

Loading examples (from DINOv3 README):

    import torch
    REPO_DIR = "/path/to/local/dinov3"
    m = torch.hub.load(REPO_DIR, 'dinov3_vits16', source='local', weights='<URL_OR_PATH>')
    # Or remote hub (auto-download default weights):
    m = torch.hub.load('facebookresearch/dinov3', 'dinov3_vits16')

This extractor samples dense patch descriptors with get_intermediate_layers(reshape=True),
then bilinearly interpolates at provided keypoints.
"""

import sys
from pathlib import Path
import torch
import torch.nn.functional as F

from easy_local_features.matching.nearest_neighbor import NearestNeighborMatcher
from omegaconf import OmegaConf
from easy_local_features.feature.basemodel import BaseExtractor, MethodType
from easy_local_features.utils import ops
from easy_local_features.utils.download import getCache
from typing import TypedDict, Optional


class DINOv3Config(TypedDict):
    weights: str
    allow_resize: bool
    repo_dir: Optional[str]
    weights_path: Optional[str]
    source: str
    normalize: str


class DINOv3_baseline(BaseExtractor):
    METHOD_TYPE = MethodType.DESCRIPTOR_ONLY
    available_weights = [
        # ViT on web images
        "dinov3_vits16",
        "dinov3_vits16plus",
        "dinov3_vitb16",
        "dinov3_vitl16",
        "dinov3_vith16plus",
        "dinov3_vit7b16",
        # ConvNeXt on web images
        "dinov3_convnext_tiny",
        "dinov3_convnext_small",
        "dinov3_convnext_base",
        "dinov3_convnext_large",
        # Satellite imagery variants (if available)
        "dinov3_sat_vitl16",
        "dinov3_sat_vit7b16",
    ]

    default_conf: DINOv3Config = {
        "weights": "dinov3_vits16",  # default ViT-S/16
        "allow_resize": True,  # resize to multiples of patch size
        # Optional advanced loading options:
        # - If you cloned the repo locally and have checkpoints, set repo_dir and weights_path.
        # - Otherwise we'll try to fetch via torch.hub (facebookresearch/dinov3) if available.
        "repo_dir": None,
        "weights_path": None,
        "source": "hub",  # "hub" or "local"
        # Normalization: 'auto' (infer from weights), 'imagenet', 'sat', or 'none'
        "normalize": "auto",
    }

    def __init__(self, conf: DINOv3Config = {}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        self.device = torch.device("cpu")
        self.matcher = NearestNeighborMatcher()

        # Upstream DINOv3 hubconf uses PEP 604 unions (e.g., float | None), which
        # require Python >= 3.10 without from __future__ import annotations.
        if sys.version_info < (3, 10):
            raise RuntimeError(
                "DINOv3 requires Python >= 3.10 due to upstream type hints (PEP 604). "
                "Please run with Python 3.10+ (e.g., uv python 3.11) or provide a patched local repo "
                "and set conf={'source': 'local', 'repo_dir': <path>, 'weights_path': <ckpt>}"
            )

        # Patch/stride size heuristic based on ViT-*/16 variants; used for grid sampling.
        self.vit_size = 16 if "16" in conf.weights else 14

        # Try to prepare a local weights path (auto-download for supported variants).
        self._maybe_prepare_weights()

        # Try to load the model. Prefer hub unless explicitly told to use local.
        self.model = None
        if conf.get("source") == "local" and conf.get("repo_dir"):
            try:
                self.model = torch.hub.load(
                    conf.repo_dir,
                    conf.weights,
                    source="local",
                    weights=conf.get("weights_path"),
                )
            except Exception:
                self.model = None

        if self.model is None:
            try:
                # Try normal hub first
                if conf.get("weights_path"):
                    self.model = torch.hub.load(
                        "facebookresearch/dinov3", conf.weights, weights=conf.get("weights_path")
                    )
                else:
                    self.model = torch.hub.load("facebookresearch/dinov3", conf.weights)
            except Exception:
                # Last-resort: force reload (helps if hub cache is stale)
                if conf.get("weights_path"):
                    self.model = torch.hub.load(
                        "facebookresearch/dinov3", conf.weights, force_reload=True, weights=conf.get("weights_path")
                    )
                else:
                    try:
                        self.model = torch.hub.load(
                            "facebookresearch/dinov3", conf.weights, force_reload=True
                        )
                    except Exception as e:
                        raise RuntimeError(
                            (
                                "Failed to load DINOv3 model. If you're using a variant without an auto-downloadable "
                                "GitHub weight, please either: (1) provide 'weights_path' to a local .pth file, or (2) "
                                "let torch.hub download the official weights by ensuring internet access.\n"
                                "For manual download instructions see: https://github.com/facebookresearch/dinov3\n"
                                f"Details: {e}"
                            )
                        )

        self.model.eval()

    def sample_features(self, keypoints, features, s=None, mode="bilinear"):
        if s is None:
            s = self.vit_size
        b, c, h, w = features.shape
        # Ensure keypoints are on the same device and dtype as features for grid_sample
        keypoints = keypoints.to(device=features.device, dtype=features.dtype)
        keypoints = keypoints / (keypoints.new_tensor([w, h]) * s)
        keypoints = keypoints * 2 - 1  # normalize to (-1, 1)
        sampled = torch.nn.functional.grid_sample(
            features, keypoints.view(b, 1, -1, 2), mode=mode, align_corners=False
        )
        sampled = torch.nn.functional.normalize(sampled.reshape(b, c, -1), p=2, dim=1)
        sampled = sampled.permute(0, 2, 1)
        return sampled

    @staticmethod
    def _apply_normalization(img: torch.Tensor, mode: str):
        """Apply per-channel normalization in-place style.
        mode: 'imagenet' | 'sat' | 'none'
        """
        if mode == "none":
            return img
        if mode == "imagenet":
            mean = torch.tensor([0.485, 0.456, 0.406], device=img.device).view(1, 3, 1, 1)
            std = torch.tensor([0.229, 0.224, 0.225], device=img.device).view(1, 3, 1, 1)
            return (img - mean) / std
        if mode == "sat":
            mean = torch.tensor([0.430, 0.411, 0.296], device=img.device).view(1, 3, 1, 1)
            std = torch.tensor([0.213, 0.156, 0.143], device=img.device).view(1, 3, 1, 1)
            return (img - mean) / std
        return img

    def denseCompute(self, img):
        # Prepare image tensor (0-1, BCHW)
        img = ops.prepareImage(img, gray=False, batch=True, imagenet=False).to(self.device)

        # Enforce 3-channel for normalization expectations
        if img.shape[1] == 1:
            img = img.repeat(1, 3, 1, 1)

        # Apply normalization per DINOv3 guidance
        norm_mode = self._infer_norm_mode()
        img = self._apply_normalization(img, norm_mode)

        if self.conf.allow_resize:
            # ensure divisible by patch size for ViT
            patch = self.vit_size
            img = F.interpolate(img, [int(x // patch * patch) for x in img.shape[-2:]])

        # Only ViT-style backbones expose get_intermediate_layers; guard for ConvNeXt
        if not hasattr(self.model, "get_intermediate_layers"):
            raise NotImplementedError(
                "DINOv3_baseline currently supports ViT backbones (dinov3_vit*)."
            )

        # DINOv3 ViT exposes get_intermediate_layers similar to DINOv2
        desc, cls_token = self.model.get_intermediate_layers(
            img, n=1, return_class_token=True, reshape=True
        )[0]

        return {
            "global_descriptor": cls_token,
            "feature_map": desc,
        }


    def _infer_norm_mode(self) -> str:
        # If explicit, honor it.
        mode = str(self.conf.get("normalize", "auto")).lower()
        if mode in {"imagenet", "sat", "none"}:
            return mode
        # Auto: inspect weights_path for 'sat493m'
        w = self.conf.get("weights_path")
        if isinstance(w, str) and "sat493m" in w.lower():
            return "sat"
        return "imagenet"

    def detectAndCompute(self, img, return_dict=None):
        raise NotImplementedError

    def detect(self, img, op=None):
        raise NotImplementedError

    def compute(self, img, keypoints=None, return_dict=False):
        # Prepare image tensor (0-1, BCHW)
        img = ops.prepareImage(img, gray=False, batch=True, imagenet=False).to(self.device)

        # Enforce 3-channel for normalization expectations
        if img.shape[1] == 1:
            img = img.repeat(1, 3, 1, 1)

        # Apply normalization per DINOv3 guidance
        norm_mode = self._infer_norm_mode()
        img = self._apply_normalization(img, norm_mode)

        if self.conf.allow_resize:
            # ensure divisible by patch size for ViT
            patch = self.vit_size
            img = F.interpolate(img, [int(x // patch * patch) for x in img.shape[-2:]])

        # Only ViT-style backbones expose get_intermediate_layers; guard for ConvNeXt
        if not hasattr(self.model, "get_intermediate_layers"):
            raise NotImplementedError(
                "DINOv3_baseline currently supports ViT backbones (dinov3_vit*)."
            )

        # DINOv3 ViT exposes get_intermediate_layers similar to DINOv2
        desc, cls_token = self.model.get_intermediate_layers(
            img, n=1, return_class_token=True, reshape=True
        )[0]

        if keypoints is None:
            raise ValueError("DINOv3_baseline.compute expects keypoints (descriptor-only method).")

        descriptors = self.sample_features(keypoints, desc, s=self.vit_size)
        if return_dict:
            return {
                "descriptors": descriptors,
                "keypoints": keypoints,
                "global_descriptor": cls_token,
                "feature_map": desc,
            }
        return keypoints, descriptors

    def to(self, device):
        self.model.to(device)
        self.device = device

    @property
    def has_detector(self):
        return False

    # ---- Internal helpers for weight resolution and auto-download ----
    @staticmethod
    def _github_release_base_url():
        # All assets hosted under the repo release tag 'dinov3'
        return (
            "https://github.com/felipecadar/easy-local-features-baselines/releases/download/dinov3/"
        )

    @staticmethod
    def _github_variant_to_filename():
        # Variants we auto-download from the project's GitHub release.
        # Keys are the torch.hub model names used in conf['weights'].
        return {
            # ViT family
            "dinov3_vits16": "dinov3_vits16_pretrain_lvd1689m-08c60483.pth",
            "dinov3_vits16plus": "dinov3_vits16plus_pretrain_lvd1689m-4057cbaa.pth",
            "dinov3_vitb16": "dinov3_vitb16_pretrain_lvd1689m-73cec8be.pth",
            "dinov3_vitl16": "dinov3_vitl16_pretrain_lvd1689m-8aa4cbdd.pth",
            "dinov3_vith16plus": "dinov3_vith16plus_pretrain_lvd1689m-7c1da9a5.pth",
            "dinov3_vit7b16": "dinov3_vit7b16_pretrain_lvd1689m-a955f4ea.pth",
            # Satellite ViT weights
            "dinov3_sat_vitl16": "dinov3_vitl16_pretrain_sat493m-eadcf0ff.pth",
            "dinov3_sat_vit7b16": "dinov3_vit7b16_pretrain_sat493m-a6675841.pth",
            # ConvNeXt family (note: compute() currently only supports ViT backbones)
            "dinov3_convnext_tiny": "dinov3_convnext_tiny_pretrain_lvd1689m-21b726bb.pth",
            "dinov3_convnext_small": "dinov3_convnext_small_pretrain_lvd1689m-296db49d.pth",
            "dinov3_convnext_base": "dinov3_convnext_base_pretrain_lvd1689m-801f2ba9.pth",
            "dinov3_convnext_large": "dinov3_convnext_large_pretrain_lvd1689m-61fa432d.pth",
        }

    def _maybe_prepare_weights(self):
        """If no weights_path is provided, try to auto-resolve/download from our GitHub release
        for supported variants. Files are saved under a temporary folder on this system.
        """
        conf = self.conf
        if conf.get("weights_path"):
            return  # user provided explicit path

        variant = conf.get("weights")
        gh_map = self._github_variant_to_filename()
        if variant not in gh_map:
            # Not one of our GitHub-hosted assets; we'll rely on torch.hub (Facebook) below.
            return

        filename = gh_map[variant]
        # Save under system temporary directory
        tmp_base = Path(getCache("dinov3")) 
        weights_dir = tmp_base
        weights_dir.mkdir(parents=True, exist_ok=True)
        target_path = weights_dir / filename

        if not target_path.exists():
            url = self._github_release_base_url() + filename
            try:
                print(f"Auto-downloading DINOv3 weights for '{variant}' from: {url}")
                torch.hub.download_url_to_file(url, str(target_path))
            except Exception as e:
                raise FileNotFoundError(
                    (
                        f"Couldn't download weights for '{variant}' to {target_path}.\n"
                        f"Tried URL: {url}\n"
                        "You may: (1) provide 'weights_path' pointing to a local .pth file, or (2) use the official\n"
                        "Facebook weights via torch.hub (internet required). See:\n"
                        "https://github.com/facebookresearch/dinov3\n"
                        f"Details: {e}"
                    )
                )

        # Use the local file we ensured above
        conf["weights_path"] = str(target_path)


if __name__ == "__main__":
    # Minimal smoke test: use SuperPoint to detect and DINOv3 to compute descriptors.
    from easy_local_features.utils import io, vis, ops
    from easy_local_features.feature.baseline_superpoint import SuperPoint_baseline

    # Load an image from the test assets
    img0 = io.fromPath("tests/assets/megadepth0.jpg")
    img1 = io.fromPath("tests/assets/megadepth1.jpg")
    
    # resize for faster testing
    img0, _ = ops.resize_short_edge(img0, 512)
    img1, _ = ops.resize_short_edge(img1, 512)
    
    # Create DINOv3 descriptor and SuperPoint detector
    # Tip: for official weights, prefer local repo + weights path per README.
    # Example:
    # method = DINOv3_baseline({
    #     "source": "local",
    #     "repo_dir": "/path/to/local/dinov3",
    #     "weights": "dinov3_vits16",
    #     "weights_path": "/path/to/dinov3_vits16_pretrain_lvd1689m-08c60483.pth",
    # })
    # For convenience we fall back to remote hub here:
    method = DINOv3_baseline(
        {
            "weights": "dinov3_vitl16",
        }
    )
    detector = SuperPoint_baseline(
        {
            "nms_radius": 4,
            "keypoint_threshold": 0.005,
            "max_keypoints": 1024,
        }
    )
    method.addDetector(detector)

    # Run detect-only, compute-only, and detect+compute for consistency
    kpts = method.detect(img0)
    kpts2, desc2 = method.detectAndCompute(img0)
    kpts3, desc3 = method.compute(img0, kpts)

    print("kpts:", kpts.shape)
    print("desc (via detectAndCompute):", desc2.shape)
    print("desc (via compute):", desc3.shape)

    # Basic sanity checks
    assert kpts.shape == kpts2.shape
    assert desc2.shape == desc3.shape
    print("DINOv3 smoke test passed.")


    matches = method.match(img0, img1)
    vis.plot_pair(
        img0, img1,
        title="DINOv3 Matches",
    )
    vis.plot_matches(
        matches["mkpts0"],
        matches["mkpts1"],
    )
    vis.show()



import torch
import os

from .pathutils import CACHE_BASE

def getCache(namespace):
    cache_path = CACHE_BASE / namespace
    if not cache_path.exists():
        os.makedirs(str(cache_path), exist_ok=True)
    return cache_path

def downloadModel(namespace, model_name, url, force=False):
    if "pth" not in model_name and "pt" not in model_name:
        model_name += ".pth"

    cache_path = getCache(namespace)
    cache_path = cache_path / model_name
    
    dirpath = cache_path.parent
    if not dirpath.exists():
        os.makedirs(str(dirpath), exist_ok=True)
    
    if not cache_path.exists() or force:
        print(f"Downloading {model_name} model... from {url}")
        torch.hub.download_url_to_file(url, str(cache_path))
    return cache_path


from pathlib import Path
from typing import Optional, Union

import numpy as np
import torch
import torch.nn as nn
from abc import ABC, abstractmethod


class Detector(ABC, nn.Module):
    @property
    @abstractmethod
    def topleft(self) -> float:
        pass

    @abstractmethod
    def load_image(im_path: Union[str, Path]) -> dict[str, torch.Tensor]:
        pass

    @abstractmethod
    def detect(
        self, batch: dict[str, torch.Tensor], *, num_keypoints, return_dense_probs=False
    ) -> dict[str, torch.Tensor]:
        pass

    @torch.inference_mode
    def detect_from_path(
        self,
        im_path: Union[str, Path],
        *,
        num_keypoints: int,
        return_dense_probs: bool = False,
    ) -> dict[str, torch.Tensor]:
        return self.detect(
            self.load_image(im_path),
            num_keypoints=num_keypoints,
            return_dense_probs=return_dense_probs,
        )

    def to_pixel_coords(
        self, normalized_coords: torch.Tensor, h: int, w: int
    ) -> torch.Tensor:
        if normalized_coords.shape[-1] != 2:
            raise ValueError(
                f"Expected shape (..., 2), but got {normalized_coords.shape}"
            )
        pixel_coords = torch.stack(
            (
                w * (normalized_coords[..., 0] + 1) / 2,
                h * (normalized_coords[..., 1] + 1) / 2,
            ),
            axis=-1,
        )
        return pixel_coords

    def to_normalized_coords(
        self, pixel_coords: torch.Tensor, h: int, w: int
    ) -> torch.Tensor:
        if pixel_coords.shape[-1] != 2:
            raise ValueError(f"Expected shape (..., 2), but got {pixel_coords.shape}")
        normalized_coords = torch.stack(
            (
                2 * (pixel_coords[..., 0]) / w - 1,
                2 * (pixel_coords[..., 1]) / h - 1,
            ),
            axis=-1,
        )
        return normalized_coords


class Matcher(ABC, nn.Module):
    @abstractmethod
    def match(
        self, im_A_path: Union[str, Path], im_B_path: Union[str, Path]
    ) -> tuple[torch.Tensor, torch.Tensor]:
        pass

    @abstractmethod
    def match_keypoints(
        self,
        keypoints_A: torch.Tensor,
        keypoints_B: torch.Tensor,
        warp: torch.Tensor,
        certainty: torch.Tensor,
        return_tuple: bool = False,
    ) -> torch.Tensor:
        pass

    @abstractmethod
    def to_pixel_coordinates(
        self, matches: torch.Tensor, h1: int, w1: int, h2: int, w2: int
    ) -> tuple[torch.Tensor, torch.Tensor]:
        pass


class Benchmark(ABC):
    def __init__(
        self,
        *,
        data_root: str,
        thresholds: list[int],
        sample_every: int = 1,
        num_ransac_runs: int = 5,
        num_keypoints: Optional[Union[list[int], int]] = None,
    ) -> None:
        self.num_keypoints = (
            [512, 1024, 2048, 4096, 8192] if num_keypoints is None else num_keypoints
        )
        if isinstance(self.num_keypoints, int):
            self.num_keypoints = [self.num_keypoints]
        self.data_root = data_root
        self.sample_every = sample_every
        self.num_ransac_runs = num_ransac_runs
        self.thresholds = thresholds

    @abstractmethod
    def benchmark(self, *, matcher: Matcher, detector: Detector) -> dict[str, float]:
        pass

    def pose_auc(self, errors):
        sort_idx = np.argsort(errors)
        errors = np.array(errors.copy())[sort_idx]
        recall = (np.arange(len(errors)) + 1) / len(errors)
        errors = np.r_[0.0, errors]
        recall = np.r_[0.0, recall]
        aucs = []
        for t in self.thresholds:
            last_index = np.searchsorted(errors, t)
            r = np.r_[recall[:last_index], recall[last_index - 1]]
            e = np.r_[errors[:last_index], t]
            aucs.append(np.trapz(r, x=e).item() / t)
        return aucs

    def compute_auc(self, errors: np.ndarray) -> dict[str, float]:
        # errors.shape = (len(benchmark)*num_keypoints*num_ransac_runs,)
        errors = (
            errors.reshape((-1, len(self.num_keypoints), self.num_ransac_runs))
            .transpose(0, 2, 1)
            .reshape(-1, len(self.num_keypoints))
        )
        results: dict[str, float] = {}
        for idx in range(len(self.num_keypoints)):
            aucs = self.pose_auc(errors[:, idx])
            for auc, th in zip(aucs, self.thresholds):
                key = (
                    f"{type(self).__name__}_auc_{th}_num_kps_{self.num_keypoints[idx]}"
                )
                results[key] = auc
        return results

    def __call__(self, *, matcher: Matcher, detector: Detector) -> dict[str, float]:
        return self.benchmark(
            matcher=matcher,
            detector=detector,
        )



import sys
from pathlib import Path

# Get the current script's directory
current_dir = Path(__file__).parent

# Add two parent directories to sys.path
parent_dir_1 = current_dir.parent.parent.parent  # Two levels up
sys.path.insert(0, str(parent_dir_1))

import warnings
import torch.nn as nn
import torch
from romatch.models.matcher import *
from romatch.models.transformer import Block, TransformerDecoder, MemEffAttention
from romatch.models.encoders import *
from romatch.models.tiny import TinyRoMa


def tiny_roma_v1_model(weights = None, freeze_xfeat=False, exact_softmax=False, xfeat = None):
    model = TinyRoMa(
        xfeat = xfeat,
        freeze_xfeat=freeze_xfeat, 
        exact_softmax=exact_softmax)
    if weights is not None:
        model.load_state_dict(weights)
    return model

def roma_model(resolution, upsample_preds, device = None, weights=None, dinov2_weights=None, amp_dtype: torch.dtype=torch.float16, **kwargs):
    # romatch weights and dinov2 weights are loaded seperately, as dinov2 weights are not parameters
    #torch.backends.cuda.matmul.allow_tf32 = True # allow tf32 on matmul TODO: these probably ruin stuff, should be careful
    #torch.backends.cudnn.allow_tf32 = True # allow tf32 on cudnn
    warnings.filterwarnings('ignore', category=UserWarning, message='TypedStorage is deprecated')
    gp_dim = 512
    feat_dim = 512
    decoder_dim = gp_dim + feat_dim
    cls_to_coord_res = 64
    coordinate_decoder = TransformerDecoder(
        nn.Sequential(*[Block(decoder_dim, 8, attn_class=MemEffAttention) for _ in range(5)]), 
        decoder_dim, 
        cls_to_coord_res**2 + 1,
        is_classifier=True,
        amp = True,
        pos_enc = False,)
    dw = True
    hidden_blocks = 8
    kernel_size = 5
    displacement_emb = "linear"
    disable_local_corr_grad = True
    
    conv_refiner = nn.ModuleDict(
        {
            "16": ConvRefiner(
                2 * 512+128+(2*7+1)**2,
                2 * 512+128+(2*7+1)**2,
                2 + 1,
                kernel_size=kernel_size,
                dw=dw,
                hidden_blocks=hidden_blocks,
                displacement_emb=displacement_emb,
                displacement_emb_dim=128,
                local_corr_radius = 7,
                corr_in_other = True,
                amp = True,
                disable_local_corr_grad = disable_local_corr_grad,
                bn_momentum = 0.01,
            ),
            "8": ConvRefiner(
                2 * 512+64+(2*3+1)**2,
                2 * 512+64+(2*3+1)**2,
                2 + 1,
                kernel_size=kernel_size,
                dw=dw,
                hidden_blocks=hidden_blocks,
                displacement_emb=displacement_emb,
                displacement_emb_dim=64,
                local_corr_radius = 3,
                corr_in_other = True,
                amp = True,
                disable_local_corr_grad = disable_local_corr_grad,
                bn_momentum = 0.01,
            ),
            "4": ConvRefiner(
                2 * 256+32+(2*2+1)**2,
                2 * 256+32+(2*2+1)**2,
                2 + 1,
                kernel_size=kernel_size,
                dw=dw,
                hidden_blocks=hidden_blocks,
                displacement_emb=displacement_emb,
                displacement_emb_dim=32,
                local_corr_radius = 2,
                corr_in_other = True,
                amp = True,
                disable_local_corr_grad = disable_local_corr_grad,
                bn_momentum = 0.01,
            ),
            "2": ConvRefiner(
                2 * 64+16,
                128+16,
                2 + 1,
                kernel_size=kernel_size,
                dw=dw,
                hidden_blocks=hidden_blocks,
                displacement_emb=displacement_emb,
                displacement_emb_dim=16,
                amp = True,
                disable_local_corr_grad = disable_local_corr_grad,
                bn_momentum = 0.01,
            ),
            "1": ConvRefiner(
                2 * 9 + 6,
                24,
                2 + 1,
                kernel_size=kernel_size,
                dw=dw,
                hidden_blocks = hidden_blocks,
                displacement_emb = displacement_emb,
                displacement_emb_dim = 6,
                amp = True,
                disable_local_corr_grad = disable_local_corr_grad,
                bn_momentum = 0.01,
            ),
        }
    )
    kernel_temperature = 0.2
    learn_temperature = False
    no_cov = True
    kernel = CosKernel
    only_attention = False
    basis = "fourier"
    gp16 = GP(
        kernel,
        T=kernel_temperature,
        learn_temperature=learn_temperature,
        only_attention=only_attention,
        gp_dim=gp_dim,
        basis=basis,
        no_cov=no_cov,
    )
    gps = nn.ModuleDict({"16": gp16})
    proj16 = nn.Sequential(nn.Conv2d(1024, 512, 1, 1), nn.BatchNorm2d(512))
    proj8 = nn.Sequential(nn.Conv2d(512, 512, 1, 1), nn.BatchNorm2d(512))
    proj4 = nn.Sequential(nn.Conv2d(256, 256, 1, 1), nn.BatchNorm2d(256))
    proj2 = nn.Sequential(nn.Conv2d(128, 64, 1, 1), nn.BatchNorm2d(64))
    proj1 = nn.Sequential(nn.Conv2d(64, 9, 1, 1), nn.BatchNorm2d(9))
    proj = nn.ModuleDict({
        "16": proj16,
        "8": proj8,
        "4": proj4,
        "2": proj2,
        "1": proj1,
        })
    displacement_dropout_p = 0.0
    gm_warp_dropout_p = 0.0
    decoder = Decoder(coordinate_decoder, 
                      gps, 
                      proj, 
                      conv_refiner, 
                      detach=True, 
                      scales=["16", "8", "4", "2", "1"], 
                      displacement_dropout_p = displacement_dropout_p,
                      gm_warp_dropout_p = gm_warp_dropout_p)
    
    encoder = CNNandDinov2(
        cnn_kwargs = dict(
            pretrained=False,
            amp = True),
        amp = True,
        use_vgg = True,
        dinov2_weights = dinov2_weights,
        amp_dtype=amp_dtype,
    )
    h,w = resolution
    symmetric = True
    attenuate_cert = True
    sample_mode = "threshold_balanced"
    matcher = RegressionMatcher(encoder, decoder, h=h, w=w, upsample_preds=upsample_preds, 
                                symmetric = symmetric, attenuate_cert = attenuate_cert, sample_mode = sample_mode, **kwargs).to(device)
    matcher.load_state_dict(weights)
    return matcher



# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb

import torch
import torch.nn as nn
import torch.nn.functional as F

from nets.sampler import *
from nets.repeatability_loss import *
from nets.reliability_loss import *


class MultiLoss (nn.Module):
    """ Combines several loss functions for convenience.
    *args: [loss weight (float), loss creator, ... ]
    
    Example:
        loss = MultiLoss( 1, MyFirstLoss(), 0.5, MySecondLoss() )
    """
    def __init__(self, *args, dbg=()):
        nn.Module.__init__(self)
        assert len(args) % 2 == 0, 'args must be a list of (float, loss)'
        self.weights = []
        self.losses = nn.ModuleList()
        for i in range(len(args)//2):
            weight = float(args[2*i+0])
            loss = args[2*i+1]
            assert isinstance(loss, nn.Module), "%s is not a loss!" % loss
            self.weights.append(weight)
            self.losses.append(loss)

    def forward(self, select=None, **variables):
        assert not select or all(1<=n<=len(self.losses) for n in select)
        d = dict()
        cum_loss = 0
        for num, (weight, loss_func) in enumerate(zip(self.weights, self.losses),1):
            if select is not None and num not in select: continue
            l = loss_func(**{k:v for k,v in variables.items()})
            if isinstance(l, tuple):
                assert len(l) == 2 and isinstance(l[1], dict)
            else:
                l = l, {loss_func.name:l}
            cum_loss = cum_loss + weight * l[0]
            for key,val in l[1].items():
                d['loss_'+key] = float(val)
        d['loss'] = float(cum_loss)
        return cum_loss, d









from typing import Dict
import numpy as np
import torch
import kornia.augmentation as K
from kornia.geometry.transform import warp_perspective

# Adapted from Kornia
class GeometricSequential:
    def __init__(self, *transforms, align_corners=True) -> None:
        self.transforms = transforms
        self.align_corners = align_corners

    def __call__(self, x, mode="bilinear"):
        b, c, h, w = x.shape
        M = torch.eye(3, device=x.device)[None].expand(b, 3, 3)
        for t in self.transforms:
            if np.random.rand() < t.p:
                M = M.matmul(
                    t.compute_transformation(x, t.generate_parameters((b, c, h, w)), None)
                )
        return (
            warp_perspective(
                x, M, dsize=(h, w), mode=mode, align_corners=self.align_corners
            ),
            M,
        )

    def apply_transform(self, x, M, mode="bilinear"):
        b, c, h, w = x.shape
        return warp_perspective(
            x, M, dsize=(h, w), align_corners=self.align_corners, mode=mode
        )


class RandomPerspective(K.RandomPerspective):
    def generate_parameters(self, batch_shape: torch.Size) -> Dict[str, torch.Tensor]:
        distortion_scale = torch.as_tensor(
            self.distortion_scale, device=self._device, dtype=self._dtype
        )
        return self.random_perspective_generator(
            batch_shape[0],
            batch_shape[-2],
            batch_shape[-1],
            distortion_scale,
            self.same_on_batch,
            self.device,
            self.dtype,
        )

    def random_perspective_generator(
        self,
        batch_size: int,
        height: int,
        width: int,
        distortion_scale: torch.Tensor,
        same_on_batch: bool = False,
        device: torch.device = torch.device("cpu"),
        dtype: torch.dtype = torch.float32,
    ) -> Dict[str, torch.Tensor]:
        r"""Get parameters for ``perspective`` for a random perspective transform.

        Args:
            batch_size (int): the tensor batch size.
            height (int) : height of the image.
            width (int): width of the image.
            distortion_scale (torch.Tensor): it controls the degree of distortion and ranges from 0 to 1.
            same_on_batch (bool): apply the same transformation across the batch. Default: False.
            device (torch.device): the device on which the random numbers will be generated. Default: cpu.
            dtype (torch.dtype): the data type of the generated random numbers. Default: float32.

        Returns:
            params Dict[str, torch.Tensor]: parameters to be passed for transformation.
                - start_points (torch.Tensor): element-wise perspective source areas with a shape of (B, 4, 2).
                - end_points (torch.Tensor): element-wise perspective target areas with a shape of (B, 4, 2).

        Note:
            The generated random numbers are not reproducible across different devices and dtypes.
        """
        if not (distortion_scale.dim() == 0 and 0 <= distortion_scale <= 1):
            raise AssertionError(
                f"'distortion_scale' must be a scalar within [0, 1]. Got {distortion_scale}."
            )
        if not (
            type(height) is int and height > 0 and type(width) is int and width > 0
        ):
            raise AssertionError(
                f"'height' and 'width' must be integers. Got {height}, {width}."
            )

        start_points: torch.Tensor = torch.tensor(
            [[[0.0, 0], [width - 1, 0], [width - 1, height - 1], [0, height - 1]]],
            device=distortion_scale.device,
            dtype=distortion_scale.dtype,
        ).expand(batch_size, -1, -1)

        # generate random offset not larger than half of the image
        fx = distortion_scale * width / 2
        fy = distortion_scale * height / 2

        factor = torch.stack([fx, fy], dim=0).view(-1, 1, 2)
        offset = (torch.rand_like(start_points) - 0.5) * 2
        end_points = start_points + factor * offset

        return dict(start_points=start_points, end_points=end_points)



class RandomErasing:
    def __init__(self, p = 0., scale = 0.) -> None:
        self.p = p
        self.scale = scale
        self.random_eraser = K.RandomErasing(scale = (0.02, scale), p = p)
    def __call__(self, image, depth):
        if self.p > 0:
            image = self.random_eraser(image)
            depth = self.random_eraser(depth, params=self.random_eraser._params)
        return image, depth
        


import warnings
import numpy as np
import cv2
import math
import torch
from torchvision import transforms
from torchvision.transforms.functional import InterpolationMode
import torch.nn.functional as F
from PIL import Image
import kornia

def recover_pose(E, kpts0, kpts1, K0, K1, mask):
    best_num_inliers = 0
    K0inv = np.linalg.inv(K0[:2,:2])
    K1inv = np.linalg.inv(K1[:2,:2])

    kpts0_n = (K0inv @ (kpts0-K0[None,:2,2]).T).T 
    kpts1_n = (K1inv @ (kpts1-K1[None,:2,2]).T).T

    for _E in np.split(E, len(E) / 3):
        n, R, t, _ = cv2.recoverPose(_E, kpts0_n, kpts1_n, np.eye(3), 1e9, mask=mask)
        if n > best_num_inliers:
            best_num_inliers = n
            ret = (R, t, mask.ravel() > 0)
    return ret



# Code taken from https://github.com/PruneTruong/DenseMatching/blob/40c29a6b5c35e86b9509e65ab0cd12553d998e5f/validation/utils_pose_estimation.py
# --- GEOMETRY ---
def estimate_pose(kpts0, kpts1, K0, K1, norm_thresh, conf=0.99999):
    if len(kpts0) < 5:
        return None
    K0inv = np.linalg.inv(K0[:2,:2])
    K1inv = np.linalg.inv(K1[:2,:2])

    kpts0 = (K0inv @ (kpts0-K0[None,:2,2]).T).T 
    kpts1 = (K1inv @ (kpts1-K1[None,:2,2]).T).T
    E, mask = cv2.findEssentialMat(
        kpts0, kpts1, np.eye(3), threshold=norm_thresh, prob=conf
    )

    ret = None
    if E is not None:
        best_num_inliers = 0

        for _E in np.split(E, len(E) / 3):
            n, R, t, _ = cv2.recoverPose(_E, kpts0, kpts1, np.eye(3), 1e9, mask=mask)
            if n > best_num_inliers:
                best_num_inliers = n
                ret = (R, t, mask.ravel() > 0)
    return ret

def estimate_pose_uncalibrated(kpts0, kpts1, K0, K1, norm_thresh, conf=0.99999):
    if len(kpts0) < 5:
        return None
    method = cv2.USAC_ACCURATE
    F, mask = cv2.findFundamentalMat(
        kpts0, kpts1, ransacReprojThreshold=norm_thresh, confidence=conf, method=method, maxIters=10000
    )
    E = K1.T@F@K0
    ret = None
    if E is not None:
        best_num_inliers = 0
        K0inv = np.linalg.inv(K0[:2,:2])
        K1inv = np.linalg.inv(K1[:2,:2])

        kpts0_n = (K0inv @ (kpts0-K0[None,:2,2]).T).T 
        kpts1_n = (K1inv @ (kpts1-K1[None,:2,2]).T).T
 
        for _E in np.split(E, len(E) / 3):
            n, R, t, _ = cv2.recoverPose(_E, kpts0_n, kpts1_n, np.eye(3), 1e9, mask=mask)
            if n > best_num_inliers:
                best_num_inliers = n
                ret = (R, t, mask.ravel() > 0)
    return ret

def unnormalize_coords(x_n,h,w):
    x = torch.stack(
        (w * (x_n[..., 0] + 1) / 2, h * (x_n[..., 1] + 1) / 2), dim=-1
    )  # [-1+1/h, 1-1/h] -> [0.5, h-0.5]
    return x


def rotate_intrinsic(K, n):
    base_rot = np.array([[0, 1, 0], [-1, 0, 0], [0, 0, 1]])
    rot = np.linalg.matrix_power(base_rot, n)
    return rot @ K


def rotate_pose_inplane(i_T_w, rot):
    rotation_matrices = [
        np.array(
            [
                [np.cos(r), -np.sin(r), 0.0, 0.0],
                [np.sin(r), np.cos(r), 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
            ],
            dtype=np.float32,
        )
        for r in [np.deg2rad(d) for d in (0, 270, 180, 90)]
    ]
    return np.dot(rotation_matrices[rot], i_T_w)


def scale_intrinsics(K, scales):
    scales = np.diag([1.0 / scales[0], 1.0 / scales[1], 1.0])
    return np.dot(scales, K)


def to_homogeneous(points):
    return np.concatenate([points, np.ones_like(points[:, :1])], axis=-1)


def angle_error_mat(R1, R2):
    cos = (np.trace(np.dot(R1.T, R2)) - 1) / 2
    cos = np.clip(cos, -1.0, 1.0)  # numercial errors can make it out of bounds
    return np.rad2deg(np.abs(np.arccos(cos)))


def angle_error_vec(v1, v2):
    n = np.linalg.norm(v1) * np.linalg.norm(v2)
    return np.rad2deg(np.arccos(np.clip(np.dot(v1, v2) / n, -1.0, 1.0)))


def compute_pose_error(T_0to1, R, t):
    R_gt = T_0to1[:3, :3]
    t_gt = T_0to1[:3, 3]
    error_t = angle_error_vec(t.squeeze(), t_gt)
    error_t = np.minimum(error_t, 180 - error_t)  # ambiguity of E estimation
    error_R = angle_error_mat(R, R_gt)
    return error_t, error_R


def pose_auc(errors, thresholds):
    sort_idx = np.argsort(errors)
    errors = np.array(errors.copy())[sort_idx]
    recall = (np.arange(len(errors)) + 1) / len(errors)
    errors = np.r_[0.0, errors]
    recall = np.r_[0.0, recall]
    aucs = []
    for t in thresholds:
        last_index = np.searchsorted(errors, t)
        r = np.r_[recall[:last_index], recall[last_index - 1]]
        e = np.r_[errors[:last_index], t]
        aucs.append(np.trapz(r, x=e) / t)
    return aucs


# From Patch2Pix https://github.com/GrumpyZhou/patch2pix
def get_depth_tuple_transform_ops_nearest_exact(resize=None):
    ops = []
    if resize:
        ops.append(TupleResizeNearestExact(resize))
    return TupleCompose(ops)

def get_depth_tuple_transform_ops(resize=None, normalize=True, unscale=False):
    ops = []
    if resize:
        ops.append(TupleResize(resize, mode=InterpolationMode.BILINEAR))
    return TupleCompose(ops)


def get_tuple_transform_ops(resize=None, normalize=True, unscale=False, clahe = False, colorjiggle_params = None):
    ops = []
    if resize:
        ops.append(TupleResize(resize))
    ops.append(TupleToTensorScaled())
    if normalize:
        ops.append(
            TupleNormalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
        )  # Imagenet mean/std
    return TupleCompose(ops)

class ToTensorScaled(object):
    """Convert a RGB PIL Image to a CHW ordered Tensor, scale the range to [0, 1]"""

    def __call__(self, im):
        if not isinstance(im, torch.Tensor):
            im = np.array(im, dtype=np.float32).transpose((2, 0, 1))
            im /= 255.0
            return torch.from_numpy(im)
        else:
            return im

    def __repr__(self):
        return "ToTensorScaled(./255)"


class TupleToTensorScaled(object):
    def __init__(self):
        self.to_tensor = ToTensorScaled()

    def __call__(self, im_tuple):
        return [self.to_tensor(im) for im in im_tuple]

    def __repr__(self):
        return "TupleToTensorScaled(./255)"


class ToTensorUnscaled(object):
    """Convert a RGB PIL Image to a CHW ordered Tensor"""

    def __call__(self, im):
        return torch.from_numpy(np.array(im, dtype=np.float32).transpose((2, 0, 1)))

    def __repr__(self):
        return "ToTensorUnscaled()"


class TupleToTensorUnscaled(object):
    """Convert a RGB PIL Image to a CHW ordered Tensor"""

    def __init__(self):
        self.to_tensor = ToTensorUnscaled()

    def __call__(self, im_tuple):
        return [self.to_tensor(im) for im in im_tuple]

    def __repr__(self):
        return "TupleToTensorUnscaled()"

class TupleResizeNearestExact:
    def __init__(self, size):
        self.size = size
    def __call__(self, im_tuple):
        return [F.interpolate(im, size = self.size, mode = 'nearest-exact') for im in im_tuple]

    def __repr__(self):
        return "TupleResizeNearestExact(size={})".format(self.size)


class TupleResize(object):
    def __init__(self, size, mode=InterpolationMode.BICUBIC):
        self.size = size
        self.resize = transforms.Resize(size, mode)
    def __call__(self, im_tuple):
        return [self.resize(im) for im in im_tuple]

    def __repr__(self):
        return "TupleResize(size={})".format(self.size)
    
class Normalize:
    def __call__(self,im):
        mean = im.mean(dim=(1,2), keepdims=True)
        std = im.std(dim=(1,2), keepdims=True)
        return (im-mean)/std


class TupleNormalize(object):
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std
        self.normalize = transforms.Normalize(mean=mean, std=std)

    def __call__(self, im_tuple):
        c,h,w = im_tuple[0].shape
        if c > 3:
            warnings.warn(f"Number of channels c={c} > 3, assuming first 3 are rgb")
        return [self.normalize(im[:3]) for im in im_tuple]

    def __repr__(self):
        return "TupleNormalize(mean={}, std={})".format(self.mean, self.std)


class TupleCompose(object):
    def __init__(self, transforms):
        self.transforms = transforms

    def __call__(self, im_tuple):
        for t in self.transforms:
            im_tuple = t(im_tuple)
        return im_tuple

    def __repr__(self):
        format_string = self.__class__.__name__ + "("
        for t in self.transforms:
            format_string += "\n"
            format_string += "    {0}".format(t)
        format_string += "\n)"
        return format_string

@torch.no_grad()
def cls_to_flow(cls, deterministic_sampling = True):
    B,C,H,W = cls.shape
    device = cls.device
    res = round(math.sqrt(C))
    G = torch.meshgrid(
        *[torch.linspace(-1+1/res, 1-1/res, steps = res, device = device) for _ in range(2)],
        indexing = 'ij'
        )
    G = torch.stack([G[1],G[0]],dim=-1).reshape(C,2)
    if deterministic_sampling:
        sampled_cls = cls.max(dim=1).indices
    else:
        sampled_cls = torch.multinomial(cls.permute(0,2,3,1).reshape(B*H*W,C).softmax(dim=-1), 1).reshape(B,H,W)
    flow = G[sampled_cls]
    return flow

@torch.no_grad()
def cls_to_flow_refine(cls):
    B,C,H,W = cls.shape
    device = cls.device
    res = round(math.sqrt(C))
    G = torch.meshgrid(
        *[torch.linspace(-1+1/res, 1-1/res, steps = res, device = device) for _ in range(2)],
        indexing = 'ij'
        )
    G = torch.stack([G[1],G[0]],dim=-1).reshape(C,2)
    # FIXME: below softmax line causes mps to bug, don't know why.
    if device.type == 'mps':
        cls = cls.log_softmax(dim=1).exp()
    else:
        cls = cls.softmax(dim=1)
    mode = cls.max(dim=1).indices
    
    index = torch.stack((mode-1, mode, mode+1, mode - res, mode + res), dim = 1).clamp(0,C - 1).long()
    neighbours = torch.gather(cls, dim = 1, index = index)[...,None]
    flow = neighbours[:,0] * G[index[:,0]] + neighbours[:,1] * G[index[:,1]] + neighbours[:,2] * G[index[:,2]] + neighbours[:,3] * G[index[:,3]] + neighbours[:,4] * G[index[:,4]]
    tot_prob = neighbours.sum(dim=1)  
    flow = flow / tot_prob
    return flow


def get_gt_warp(depth1, depth2, T_1to2, K1, K2, depth_interpolation_mode = 'bilinear', relative_depth_error_threshold = 0.05, H = None, W = None):
    
    if H is None:
        B,H,W = depth1.shape
    else:
        B = depth1.shape[0]
    with torch.no_grad():
        x1_n = torch.meshgrid(
            *[
                torch.linspace(
                    -1 + 1 / n, 1 - 1 / n, n, device=depth1.device
                )
                for n in (B, H, W)
            ],
            indexing = 'ij'
        )
        x1_n = torch.stack((x1_n[2], x1_n[1]), dim=-1).reshape(B, H * W, 2)
        mask, x2 = warp_kpts(
            x1_n.double(),
            depth1.double(),
            depth2.double(),
            T_1to2.double(),
            K1.double(),
            K2.double(),
            depth_interpolation_mode = depth_interpolation_mode,
            relative_depth_error_threshold = relative_depth_error_threshold,
        )
        prob = mask.float().reshape(B, H, W)
        x2 = x2.reshape(B, H, W, 2)
        return x2, prob

@torch.no_grad()
def warp_kpts(kpts0, depth0, depth1, T_0to1, K0, K1, smooth_mask = False, return_relative_depth_error = False, depth_interpolation_mode = "bilinear", relative_depth_error_threshold = 0.05):
    """Warp kpts0 from I0 to I1 with depth, K and Rt
    Also check covisibility and depth consistency.
    Depth is consistent if relative error < 0.2 (hard-coded).
    # https://github.com/zju3dv/LoFTR/blob/94e98b695be18acb43d5d3250f52226a8e36f839/src/loftr/utils/geometry.py adapted from here
    Args:
        kpts0 (torch.Tensor): [N, L, 2] - <x, y>, should be normalized in (-1,1)
        depth0 (torch.Tensor): [N, H, W],
        depth1 (torch.Tensor): [N, H, W],
        T_0to1 (torch.Tensor): [N, 3, 4],
        K0 (torch.Tensor): [N, 3, 3],
        K1 (torch.Tensor): [N, 3, 3],
    Returns:
        calculable_mask (torch.Tensor): [N, L]
        warped_keypoints0 (torch.Tensor): [N, L, 2] <x0_hat, y1_hat>
    """
    (
        n,
        h,
        w,
    ) = depth0.shape
    if depth_interpolation_mode == "combined":
        # Inspired by approach in inloc, try to fill holes from bilinear interpolation by nearest neighbour interpolation
        if smooth_mask:
            raise NotImplementedError("Combined bilinear and NN warp not implemented")
        valid_bilinear, warp_bilinear = warp_kpts(kpts0, depth0, depth1, T_0to1, K0, K1, 
                  smooth_mask = smooth_mask, 
                  return_relative_depth_error = return_relative_depth_error, 
                  depth_interpolation_mode = "bilinear",
                  relative_depth_error_threshold = relative_depth_error_threshold)
        valid_nearest, warp_nearest = warp_kpts(kpts0, depth0, depth1, T_0to1, K0, K1, 
                  smooth_mask = smooth_mask, 
                  return_relative_depth_error = return_relative_depth_error, 
                  depth_interpolation_mode = "nearest-exact",
                  relative_depth_error_threshold = relative_depth_error_threshold)
        nearest_valid_bilinear_invalid = (~valid_bilinear).logical_and(valid_nearest) 
        warp = warp_bilinear.clone()
        warp[nearest_valid_bilinear_invalid] = warp_nearest[nearest_valid_bilinear_invalid]
        valid = valid_bilinear | valid_nearest
        return valid, warp
        
        
    kpts0_depth = F.grid_sample(depth0[:, None], kpts0[:, :, None], mode = depth_interpolation_mode, align_corners=False)[
        :, 0, :, 0
    ]
    kpts0 = torch.stack(
        (w * (kpts0[..., 0] + 1) / 2, h * (kpts0[..., 1] + 1) / 2), dim=-1
    )  # [-1+1/h, 1-1/h] -> [0.5, h-0.5]
    # Sample depth, get calculable_mask on depth != 0
    nonzero_mask = kpts0_depth != 0

    # Unproject
    kpts0_h = (
        torch.cat([kpts0, torch.ones_like(kpts0[:, :, [0]])], dim=-1)
        * kpts0_depth[..., None]
    )  # (N, L, 3)
    kpts0_n = K0.inverse() @ kpts0_h.transpose(2, 1)  # (N, 3, L)
    kpts0_cam = kpts0_n

    # Rigid Transform
    w_kpts0_cam = T_0to1[:, :3, :3] @ kpts0_cam + T_0to1[:, :3, [3]]  # (N, 3, L)
    w_kpts0_depth_computed = w_kpts0_cam[:, 2, :]

    # Project
    w_kpts0_h = (K1 @ w_kpts0_cam).transpose(2, 1)  # (N, L, 3)
    w_kpts0 = w_kpts0_h[:, :, :2] / (
        w_kpts0_h[:, :, [2]] + 1e-4
    )  # (N, L, 2), +1e-4 to avoid zero depth

    # Covisible Check
    h, w = depth1.shape[1:3]
    covisible_mask = (
        (w_kpts0[:, :, 0] > 0)
        * (w_kpts0[:, :, 0] < w - 1)
        * (w_kpts0[:, :, 1] > 0)
        * (w_kpts0[:, :, 1] < h - 1)
    )
    w_kpts0 = torch.stack(
        (2 * w_kpts0[..., 0] / w - 1, 2 * w_kpts0[..., 1] / h - 1), dim=-1
    )  # from [0.5,h-0.5] -> [-1+1/h, 1-1/h]
    # w_kpts0[~covisible_mask, :] = -5 # xd

    w_kpts0_depth = F.grid_sample(
        depth1[:, None], w_kpts0[:, :, None], mode=depth_interpolation_mode, align_corners=False
    )[:, 0, :, 0]
    
    relative_depth_error = (
        (w_kpts0_depth - w_kpts0_depth_computed) / w_kpts0_depth
    ).abs()
    if not smooth_mask:
        consistent_mask = relative_depth_error < relative_depth_error_threshold
    else:
        consistent_mask = (-relative_depth_error/smooth_mask).exp()
    valid_mask = nonzero_mask * covisible_mask * consistent_mask
    if return_relative_depth_error:
        return relative_depth_error, w_kpts0
    else:
        return valid_mask, w_kpts0

imagenet_mean = torch.tensor([0.485, 0.456, 0.406])
imagenet_std = torch.tensor([0.229, 0.224, 0.225])


def numpy_to_pil(x: np.ndarray):
    """
    Args:
        x: Assumed to be of shape (h,w,c)
    """
    if isinstance(x, torch.Tensor):
        x = x.detach().cpu().numpy()
    if x.max() <= 1.01:
        x *= 255
    x = x.astype(np.uint8)
    return Image.fromarray(x)


def tensor_to_pil(x, unnormalize=False):
    if unnormalize:
        x = x * (imagenet_std[:, None, None].to(x.device)) + (imagenet_mean[:, None, None].to(x.device))
    x = x.detach().permute(1, 2, 0).cpu().numpy()
    x = np.clip(x, 0.0, 1.0)
    return numpy_to_pil(x)


def to_cuda(batch):
    for key, value in batch.items():
        if isinstance(value, torch.Tensor):
            batch[key] = value.cuda()
    return batch


def to_cpu(batch):
    for key, value in batch.items():
        if isinstance(value, torch.Tensor):
            batch[key] = value.cpu()
    return batch


def get_pose(calib):
    w, h = np.array(calib["imsize"])[0]
    return np.array(calib["K"]), np.array(calib["R"]), np.array(calib["T"]).T, h, w


def compute_relative_pose(R1, t1, R2, t2):
    rots = R2 @ (R1.T)
    trans = -rots @ t1 + t2
    return rots, trans

@torch.no_grad()
def reset_opt(opt):
    for group in opt.param_groups:
        for p in group['params']:
            if p.requires_grad:
                state = opt.state[p]
                # State initialization

                # Exponential moving average of gradient values
                state['exp_avg'] = torch.zeros_like(p)
                # Exponential moving average of squared gradient values
                state['exp_avg_sq'] = torch.zeros_like(p)
                # Exponential moving average of gradient difference
                state['exp_avg_diff'] = torch.zeros_like(p)


def flow_to_pixel_coords(flow, h1, w1):
    flow = (
        torch.stack(
            (
                w1 * (flow[..., 0] + 1) / 2,
                h1 * (flow[..., 1] + 1) / 2,
            ),
            axis=-1,
        )
    )
    return flow

to_pixel_coords = flow_to_pixel_coords # just an alias

def flow_to_normalized_coords(flow, h1, w1):
    flow = (
        torch.stack(
            (
                2 * (flow[..., 0]) / w1 - 1,
                2 * (flow[..., 1]) / h1 - 1,
            ),
            axis=-1,
        )
    )
    return flow

to_normalized_coords = flow_to_normalized_coords # just an alias

def warp_to_pixel_coords(warp, h1, w1, h2, w2):
    warp1 = warp[..., :2]
    warp1 = (
        torch.stack(
            (
                w1 * (warp1[..., 0] + 1) / 2,
                h1 * (warp1[..., 1] + 1) / 2,
            ),
            axis=-1,
        )
    )
    warp2 = warp[..., 2:]
    warp2 = (
        torch.stack(
            (
                w2 * (warp2[..., 0] + 1) / 2,
                h2 * (warp2[..., 1] + 1) / 2,
            ),
            axis=-1,
        )
    )
    return torch.cat((warp1,warp2), dim=-1)



def signed_point_line_distance(point, line, eps: float = 1e-9):
    r"""Return the distance from points to lines.

    Args:
       point: (possibly homogeneous) points :math:`(*, N, 2 or 3)`.
       line: lines coefficients :math:`(a, b, c)` with shape :math:`(*, N, 3)`, where :math:`ax + by + c = 0`.
       eps: Small constant for safe sqrt.

    Returns:
        the computed distance with shape :math:`(*, N)`.
    """

    if not point.shape[-1] in (2, 3):
        raise ValueError(f"pts must be a (*, 2 or 3) tensor. Got {point.shape}")

    if not line.shape[-1] == 3:
        raise ValueError(f"lines must be a (*, 3) tensor. Got {line.shape}")

    numerator = (line[..., 0] * point[..., 0] + line[..., 1] * point[..., 1] + line[..., 2])
    denominator = line[..., :2].norm(dim=-1)

    return numerator / (denominator + eps)


def signed_left_to_right_epipolar_distance(pts1, pts2, Fm):
    r"""Return one-sided epipolar distance for correspondences given the fundamental matrix.

    This method measures the distance from points in the right images to the epilines
    of the corresponding points in the left images as they reflect in the right images.

    Args:
       pts1: correspondences from the left images with shape
         :math:`(*, N, 2 or 3)`. If they are not homogeneous, converted automatically.
       pts2: correspondences from the right images with shape
         :math:`(*, N, 2 or 3)`. If they are not homogeneous, converted automatically.
       Fm: Fundamental matrices with shape :math:`(*, 3, 3)`. Called Fm to
         avoid ambiguity with torch.nn.functional.

    Returns:
        the computed Symmetrical distance with shape :math:`(*, N)`.
    """
    import kornia
    if (len(Fm.shape) < 3) or not Fm.shape[-2:] == (3, 3):
        raise ValueError(f"Fm must be a (*, 3, 3) tensor. Got {Fm.shape}")

    if pts1.shape[-1] == 2:
        pts1 = kornia.geometry.convert_points_to_homogeneous(pts1)

    F_t = Fm.transpose(dim0=-2, dim1=-1)
    line1_in_2 = pts1 @ F_t

    return signed_point_line_distance(pts2, line1_in_2)

def get_grid(b, h, w, device):
    grid = torch.meshgrid(
        *[
            torch.linspace(-1 + 1 / n, 1 - 1 / n, n, device=device)
            for n in (b, h, w)
        ],
        indexing = 'ij'
    )
    grid = torch.stack((grid[2], grid[1]), dim=-1).reshape(b, h, w, 2)
    return grid


def get_autocast_params(device=None, enabled=False, dtype=None):
    if device is None:
        autocast_device = "cuda" if torch.cuda.is_available() else "cpu"
    else:
        #strip :X from device
        autocast_device = str(device).split(":")[0]
    if 'cuda' in str(device):
        out_dtype = dtype
        enabled = True
    else:
        out_dtype = torch.bfloat16
        enabled = False
        # mps is not supported
        autocast_device = "cpu"
    return autocast_device, enabled, out_dtype

def check_not_i16(im):
    if im.mode == "I;16":
        raise NotImplementedError("Can't handle 16 bit images")

def check_rgb(im):
    if im.mode != "RGB":
        raise NotImplementedError("Can't handle non-RGB images")



# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb
import numpy as np
from PIL import Image, ImageOps, ImageEnhance


class DummyImg:
    ''' This class is a dummy image only defined by its size.
    '''
    def __init__(self, size):
        self.size = size
        
    def resize(self, size, *args, **kwargs):
        return DummyImg(size)
        
    def expand(self, border):
        w, h = self.size
        if isinstance(border, int):
            size = (w+2*border, h+2*border)
        else:
            l,t,r,b = border
            size = (w+l+r, h+t+b)
        return DummyImg(size)

    def crop(self, border):
        w, h = self.size
        l,t,r,b = border
        assert 0 <= l <= r <= w
        assert 0 <= t <= b <= h
        size = (r-l, b-t)
        return DummyImg(size)
    
    def rotate(self, angle):
        raise NotImplementedError

    def transform(self, size, *args, **kwargs):
        return DummyImg(size)


def grab_img( img_and_label ):
    ''' Called to extract the image from an img_and_label input
    (a dictionary). Also compatible with old-style PIL images.
    '''
    if isinstance(img_and_label, dict):
        # if input is a dictionary, then
        # it must contains the img or its size.
        try:
            return img_and_label['img']
        except KeyError:
            return DummyImg(img_and_label['imsize'])
            
    else:
        # or it must be the img directly
        return img_and_label


def update_img_and_labels(img_and_label, img, persp=None):
    ''' Called to update the img_and_label
    '''
    if isinstance(img_and_label, dict):
        img_and_label['img'] = img
        img_and_label['imsize'] = img.size

        if persp:
            if 'persp' not in img_and_label:
                img_and_label['persp'] = (1,0,0,0,1,0,0,0)
            img_and_label['persp'] = persp_mul(persp, img_and_label['persp'])
        
        return img_and_label
        
    else:
        # or it must be the img directly
        return img


def rand_log_uniform(a, b):
    return np.exp(np.random.uniform(np.log(a),np.log(b)))


def translate(tx, ty):
    return (1,0,tx,
            0,1,ty,
            0,0)

def rotate(angle):
    return (np.cos(angle),-np.sin(angle), 0,
            np.sin(angle), np.cos(angle), 0,
            0, 0)


def persp_mul(mat, mat2):
    ''' homography (perspective) multiplication.
    mat: 8-tuple (homography transform)
    mat2: 8-tuple (homography transform) or 2-tuple (point)
    '''
    assert isinstance(mat, tuple)
    assert isinstance(mat2, tuple)

    mat = np.float32(mat+(1,)).reshape(3,3)
    mat2 = np.array(mat2+(1,)).reshape(3,3)
    res = np.dot(mat, mat2)
    return tuple((res/res[2,2]).ravel()[:8])


def persp_apply(mat, pts):
    ''' homography (perspective) transformation.
    mat: 8-tuple (homography transform)
    pts: numpy array
    '''
    assert isinstance(mat, tuple)
    assert isinstance(pts, np.ndarray)
    assert pts.shape[-1] == 2
    mat = np.float32(mat+(1,)).reshape(3,3)

    if pts.ndim == 1:
        pt = np.dot(pts, mat[:,:2].T).ravel() + mat[:,2]
        pt /= pt[2] # homogeneous coordinates
        return tuple(pt[:2])
    else:
        pt = np.dot(pts, mat[:,:2].T) + mat[:,2]
        pt[:,:2] /= pt[:,2:3] # homogeneous coordinates
        return pt[:,:2]


def is_pil_image(img):
    return isinstance(img, Image.Image)


def adjust_brightness(img, brightness_factor):
    """Adjust brightness of an Image.
    Args:
    img (PIL Image): PIL Image to be adjusted.
    brightness_factor (float):  How much to adjust the brightness. Can be
    any non negative number. 0 gives a black image, 1 gives the
    original image while 2 increases the brightness by a factor of 2.
    Returns:
    PIL Image: Brightness adjusted image.
    Copied from https://github.com/pytorch in torchvision/transforms/functional.py
    """
    if not is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    enhancer = ImageEnhance.Brightness(img)
    img = enhancer.enhance(brightness_factor)
    return img


def adjust_contrast(img, contrast_factor):
    """Adjust contrast of an Image.
    Args:
    img (PIL Image): PIL Image to be adjusted.
    contrast_factor (float): How much to adjust the contrast. Can be any
    non negative number. 0 gives a solid gray image, 1 gives the
    original image while 2 increases the contrast by a factor of 2.
    Returns:
    PIL Image: Contrast adjusted image.
    Copied from https://github.com/pytorch in torchvision/transforms/functional.py
    """
    if not is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    enhancer = ImageEnhance.Contrast(img)
    img = enhancer.enhance(contrast_factor)
    return img


def adjust_saturation(img, saturation_factor):
    """Adjust color saturation of an image.
    Args:
    img (PIL Image): PIL Image to be adjusted.
    saturation_factor (float):  How much to adjust the saturation. 0 will
    give a black and white image, 1 will give the original image while
    2 will enhance the saturation by a factor of 2.
    Returns:
    PIL Image: Saturation adjusted image.
    Copied from https://github.com/pytorch in torchvision/transforms/functional.py
    """
    if not is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    enhancer = ImageEnhance.Color(img)
    img = enhancer.enhance(saturation_factor)
    return img


def adjust_hue(img, hue_factor):
    """Adjust hue of an image.
    The image hue is adjusted by converting the image to HSV and
    cyclically shifting the intensities in the hue channel (H).
    The image is then converted back to original image mode.
    `hue_factor` is the amount of shift in H channel and must be in the
    interval `[-0.5, 0.5]`.
    See https://en.wikipedia.org/wiki/Hue for more details on Hue.
    Args:
    img (PIL Image): PIL Image to be adjusted.
    hue_factor (float):  How much to shift the hue channel. Should be in
    [-0.5, 0.5]. 0.5 and -0.5 give complete reversal of hue channel in
    HSV space in positive and negative direction respectively.
    0 means no shift. Therefore, both -0.5 and 0.5 will give an image
    with complementary colors while 0 gives the original image.
    Returns:
    PIL Image: Hue adjusted image.
    Copied from https://github.com/pytorch in torchvision/transforms/functional.py
    """
    if not(-0.5 <= hue_factor <= 0.5):
        raise ValueError('hue_factor is not in [-0.5, 0.5].'.format(hue_factor))

    if not is_pil_image(img):
        raise TypeError('img should be PIL Image. Got {}'.format(type(img)))

    input_mode = img.mode
    if input_mode in {'L', '1', 'I', 'F'}:
        return img

    h, s, v = img.convert('HSV').split()

    np_h = np.array(h, dtype=np.uint8)
    # uint8 addition take cares of rotation across boundaries
    with np.errstate(over='ignore'):
        np_h += np.uint8(hue_factor * 255)
        h = Image.fromarray(np_h, 'L')

    img = Image.merge('HSV', (h, s, v)).convert(input_mode)
    return img






## Spair71K dataset
import torch
import os
from omegaconf import OmegaConf


SPAIR_URL = "https://cvlab.postech.ac.kr/research/SPair-71k/data/SPair-71k.tar.gz"

from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torchvision import transforms
from PIL import Image
import numpy as np
import torch
import glob
import json
import os

def download_spair71k(root_dir):
    if os.path.exists(root_dir + "/SPair-71k"):
        print("Dataset already downloaded.")
        return
    print("Downloading SPair-71k dataset...")
    # create the directory if it does not exist
    os.makedirs(root_dir, exist_ok=True)
    # download the dataset        
    torch.hub.download_url_to_file(SPAIR_URL, root_dir + "/SPair-71k.tar.gz")
    import tarfile
    with tarfile.open(root_dir + "/SPair-71k.tar.gz", "r:gz") as tar:
        tar.extractall(path=root_dir)
    # remove the tar file
    os.remove(root_dir + "/SPair-71k.tar.gz")
    print("Download complete.")

class Normalize(object):
    def __init__(self, image_keys):
        self.image_keys = image_keys
        self.normalize = transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])

    def __call__(self, image):
        for key in self.image_keys:
            image[key] /= 255.0
            image[key] = self.normalize(image[key])
        return image


def read_img(path):
    img = np.array(Image.open(path).convert('RGB'))

    return torch.tensor(img.transpose(2, 0, 1).astype(np.float32))


class SPairDataset(Dataset):
    '''SPair-71k dataset class.
    Args dict:
        conf = {
            root_dir (str): Root directory of the dataset.
            download (bool): If True, download the dataset.
            dataset_size (str): Size of the dataset. Options: 'large', 'small'.
            split (str): Split of the dataset. Options: 'train', 'val', 'test'.
            pck_alpha (float): PCK alpha value.
        }
    '''
    
    default_conf = {
        "root_dir": "data/spair71k",
        "download": True,
        "dataset_size": "large",
        "split": "train",
        "pck_alpha": 0.1,
        "shuffle": True,
    }
    
    def __init__(self, conf={}):
        self.conf = conf = OmegaConf.merge(OmegaConf.create(self.default_conf), conf)
        
        root_dir = conf.root_dir
        if conf.download:
            download_spair71k(root_dir)
            
        self.images_dir = os.path.join(root_dir, "SPair-71k", "JPEGImages")
        self.annotations_dir = os.path.join(root_dir, "SPair-71k", "PairAnnotation")
        self.segmentation_dir = os.path.join(root_dir, "SPair-71k", "Segmentation")
        self.layout_dir = os.path.join(root_dir, "SPair-71k", "Layout")
        self.image_annotations_dir = os.path.join(root_dir, "SPair-71k", "ImageAnnotation")
        
        self.datatype = conf.split
        if self.datatype == 'train':
            self.datatype = 'trn'
        
        self.pck_alpha = conf.pck_alpha
        self.ann_files = open(os.path.join(self.layout_dir, conf.dataset_size, self.datatype + '.txt'), "r").read().split('\n')
        self.ann_files = self.ann_files[:len(self.ann_files) - 1]
        self.pair_ann_path = self.annotations_dir
        self.image_path = self.images_dir
        self.categories = list(map(lambda x: os.path.basename(x), glob.glob('%s/*' % self.image_path)))
        self.categories.sort()
        # self.transform = Normalize(['src_img', 'trg_img'])
        self.transform = None
        
        if conf.shuffle:
            np.random.shuffle(self.ann_files)
        print(f"Loaded {len(self.ann_files)} {self.datatype} samples from {self.image_path}")

    def __len__(self):
        return len(self.ann_files)

    def __getitem__(self, idx):
        # get pre-processed images
        ann_file = self.ann_files[idx] + '.json'
        with open(os.path.join(self.pair_ann_path, self.datatype, ann_file)) as f:
            annotation = json.load(f)

        category = annotation['category']
        src_img = read_img(os.path.join(self.image_path, category, annotation['src_imname']))
        trg_img = read_img(os.path.join(self.image_path, category, annotation['trg_imname']))

        trg_bbox = annotation['trg_bndbox']
        pck_threshold = max(trg_bbox[2] - trg_bbox[0],  trg_bbox[3] - trg_bbox[1]) * self.pck_alpha

        sample = {'pair_id': annotation['pair_id'],
                  'filename': annotation['filename'],
                  'src_imname': annotation['src_imname'],
                  'trg_imname': annotation['trg_imname'],
                  'src_imsize': src_img.size(),
                  'trg_imsize': trg_img.size(),

                  'src_bbox': annotation['src_bndbox'],
                  'trg_bbox': annotation['trg_bndbox'],
                  'category': annotation['category'],

                  'src_pose': annotation['src_pose'],
                  'trg_pose': annotation['trg_pose'],

                  'src_img': src_img,
                  'trg_img': trg_img,
                  'src_kps': torch.tensor(annotation['src_kps']).float(),
                  'trg_kps': torch.tensor(annotation['trg_kps']).float(),

                  'mirror': annotation['mirror'],
                  'vp_var': annotation['viewpoint_variation'],
                  'sc_var': annotation['scale_variation'],
                  'truncn': annotation['truncation'],
                  'occlsn': annotation['occlusion'],

                  'pck_threshold': pck_threshold}

        if self.transform:
            sample = self.transform(sample)

        return sample


# if __name__ == '__main__':
#     pair_ann_path = '../PairAnnotation'
#     layout_path = '../Layout'
#     image_path = '../JPEGImages'
#     dataset_size = 'large'
#     pck_alpha = 0.1

#     trn_dataset = SPairDataset(pair_ann_path, layout_path, image_path, dataset_size, pck_alpha, datatype='trn')
#     val_dataset = SPairDataset(pair_ann_path, layout_path, image_path, dataset_size, pck_alpha, datatype='val')
#     test_dataset = SPairDataset(pair_ann_path, layout_path, image_path, dataset_size, pck_alpha, datatype='test')

#     trn_dataloader = DataLoader(trn_dataset, num_workers=1)
#     val_dataloader = DataLoader(val_dataset, num_workers=1)
#     test_dataloader = DataLoader(test_dataset, num_workers=1)

#     for data in trn_dataloader:
#         print('Trn: %20s %20s %10.2f' % (data['src_imname'], data['trg_imname'], data['pck_threshold']))

#     for data in val_dataloader:
#         print('Val: %20s %20s %10.2f' % (data['src_imname'], data['trg_imname'], data['pck_threshold']))

#     for data in test_dataloader:
#         print('Test: %20s %20s %10.2f' % (data['src_imname'], data['trg_imname'], data['pck_threshold']))

#     print('SPair-71k dataset implementation example finished.')


if __name__ == "__main__":
    
    from easy_local_features.utils import io, vis, ops
    
    spair = SPairDataset({
        "root_dir": "data/spair71k",
        "download": True,
        "dataset_size": "large",
        "split": "val",
        "pck_alpha": 0.1,
    })
    
    for data in spair:
        im1 = data['src_img']
        im2 = data['trg_img']
        
        kpts1 = data['src_kps']
        kpts2 = data['trg_kps']
        
        
        vis.plot_pair(im1, im2)
        vis.plot_matches(kpts1, kpts2)
        vis.show()


# Copyright 2019-present NAVER Corp.
# CC BY-NC-SA 3.0
# Available only for non-commercial use

import pdb
import torch.nn as nn
import torch.nn.functional as F

from nets.ap_loss import APLoss


class PixelAPLoss (nn.Module):
    """ Computes the pixel-wise AP loss:
        Given two images and ground-truth optical flow, computes the AP per pixel.
        
        feat1:  (B, C, H, W)   pixel-wise features extracted from img1
        feat2:  (B, C, H, W)   pixel-wise features extracted from img2
        aflow:  (B, 2, H, W)   absolute flow: aflow[...,y1,x1] = x2,y2
    """
    def __init__(self, sampler, nq=20):
        nn.Module.__init__(self)
        self.aploss = APLoss(nq, min=0, max=1, euc=False)
        self.name = 'pixAP'
        self.sampler = sampler

    def loss_from_ap(self, ap, rel):
        return 1 - ap

    def forward(self, descriptors, aflow, **kw):
        # subsample things
        scores, gt, msk, qconf = self.sampler(descriptors, kw.get('reliability'), aflow)
        
        # compute pixel-wise AP
        n = qconf.numel()
        if n == 0: return 0
        scores, gt = scores.view(n,-1), gt.view(n,-1)
        ap = self.aploss(scores, gt).view(msk.shape)

        pixel_loss = self.loss_from_ap(ap, qconf)
        
        loss = pixel_loss[msk].mean()
        return loss


class ReliabilityLoss (PixelAPLoss):
    """ same than PixelAPLoss, but also train a pixel-wise confidence
        that this pixel is going to have a good AP.
    """
    def __init__(self, sampler, base=0.5, **kw):
        PixelAPLoss.__init__(self, sampler, **kw)
        assert 0 <= base < 1
        self.base = base
        self.name = 'reliability'

    def loss_from_ap(self, ap, rel):
        return 1 - ap*rel - (1-rel)*self.base






# %BANNER_BEGIN%
# ---------------------------------------------------------------------
# %COPYRIGHT_BEGIN%
#
#  Magic Leap, Inc. ("COMPANY") CONFIDENTIAL
#
#  Unpublished Copyright (c) 2020
#  Magic Leap, Inc., All Rights Reserved.
#
# NOTICE:  All information contained herein is, and remains the property
# of COMPANY. The intellectual and technical concepts contained herein
# are proprietary to COMPANY and may be covered by U.S. and Foreign
# Patents, patents in process, and are protected by trade secret or
# copyright law.  Dissemination of this information or reproduction of
# this material is strictly forbidden unless prior written permission is
# obtained from COMPANY.  Access to the source code contained herein is
# hereby forbidden to anyone except current COMPANY employees, managers
# or contractors who have executed Confidentiality and Non-disclosure
# agreements explicitly covering such access.
#
# The copyright notice above does not evidence any actual or intended
# publication or disclosure  of  this source code, which includes
# information that is confidential and/or proprietary, and is a trade
# secret, of  COMPANY.   ANY REPRODUCTION, MODIFICATION, DISTRIBUTION,
# PUBLIC  PERFORMANCE, OR PUBLIC DISPLAY OF OR THROUGH USE  OF THIS
# SOURCE CODE  WITHOUT THE EXPRESS WRITTEN CONSENT OF COMPANY IS
# STRICTLY PROHIBITED, AND IN VIOLATION OF APPLICABLE LAWS AND
# INTERNATIONAL TREATIES.  THE RECEIPT OR POSSESSION OF  THIS SOURCE
# CODE AND/OR RELATED INFORMATION DOES NOT CONVEY OR IMPLY ANY RIGHTS
# TO REPRODUCE, DISCLOSE OR DISTRIBUTE ITS CONTENTS, OR TO MANUFACTURE,
# USE, OR SELL ANYTHING THAT IT  MAY DESCRIBE, IN WHOLE OR IN PART.
#
# %COPYRIGHT_END%
# ----------------------------------------------------------------------
# %AUTHORS_BEGIN%
#
#  Originating Authors: Paul-Edouard Sarlin
#
# %AUTHORS_END%
# --------------------------------------------------------------------*/
# %BANNER_END%

from copy import deepcopy
from pathlib import Path
from typing import List, Tuple

import torch
from torch import nn


def MLP(channels: List[int], do_bn: bool = True) -> nn.Module:
    """ Multi-layer perceptron """
    n = len(channels)
    layers = []
    for i in range(1, n):
        layers.append(
            nn.Conv1d(channels[i - 1], channels[i], kernel_size=1, bias=True))
        if i < (n-1):
            if do_bn:
                layers.append(nn.BatchNorm1d(channels[i]))
            layers.append(nn.ReLU())
    return nn.Sequential(*layers)


def normalize_keypoints(kpts, image_shape):
    """ Normalize keypoints locations based on image image_shape"""
    _, _, height, width = image_shape
    one = kpts.new_tensor(1)
    size = torch.stack([one*width, one*height])[None]
    center = size / 2
    scaling = size.max(1, keepdim=True).values * 0.7
    return (kpts - center[:, None, :]) / scaling[:, None, :]


class KeypointEncoder(nn.Module):
    """ Joint encoding of visual appearance and location using MLPs"""
    def __init__(self, feature_dim: int, layers: List[int]) -> None:
        super().__init__()
        self.encoder = MLP([3] + layers + [feature_dim])
        nn.init.constant_(self.encoder[-1].bias, 0.0)

    def forward(self, kpts, scores):
        inputs = [kpts.transpose(1, 2), scores.unsqueeze(1)]
        return self.encoder(torch.cat(inputs, dim=1))


def attention(query: torch.Tensor, key: torch.Tensor, value: torch.Tensor) -> Tuple[torch.Tensor,torch.Tensor]:
    dim = query.shape[1]
    scores = torch.einsum('bdhn,bdhm->bhnm', query, key) / dim**.5
    prob = torch.nn.functional.softmax(scores, dim=-1)
    return torch.einsum('bhnm,bdhm->bdhn', prob, value), prob


class MultiHeadedAttention(nn.Module):
    """ Multi-head attention to increase model expressivitiy """
    def __init__(self, num_heads: int, d_model: int):
        super().__init__()
        assert d_model % num_heads == 0
        self.dim = d_model // num_heads
        self.num_heads = num_heads
        self.merge = nn.Conv1d(d_model, d_model, kernel_size=1)
        self.proj = nn.ModuleList([deepcopy(self.merge) for _ in range(3)])

    def forward(self, query: torch.Tensor, key: torch.Tensor, value: torch.Tensor) -> torch.Tensor:
        batch_dim = query.size(0)
        query, key, value = [l(x).view(batch_dim, self.dim, self.num_heads, -1)
                             for l, x in zip(self.proj, (query, key, value))]
        x, _ = attention(query, key, value)
        return self.merge(x.contiguous().view(batch_dim, self.dim*self.num_heads, -1))


class AttentionalPropagation(nn.Module):
    def __init__(self, feature_dim: int, num_heads: int):
        super().__init__()
        self.attn = MultiHeadedAttention(num_heads, feature_dim)
        self.mlp = MLP([feature_dim*2, feature_dim*2, feature_dim])
        nn.init.constant_(self.mlp[-1].bias, 0.0)

    def forward(self, x: torch.Tensor, source: torch.Tensor) -> torch.Tensor:
        message = self.attn(x, source, source)
        return self.mlp(torch.cat([x, message], dim=1))


class AttentionalGNN(nn.Module):
    def __init__(self, feature_dim: int, layer_names: List[str]) -> None:
        super().__init__()
        self.layers = nn.ModuleList([
            AttentionalPropagation(feature_dim, 4)
            for _ in range(len(layer_names))])
        self.names = layer_names

    def forward(self, desc0: torch.Tensor, desc1: torch.Tensor) -> Tuple[torch.Tensor,torch.Tensor]:
        for layer, name in zip(self.layers, self.names):
            if name == 'cross':
                src0, src1 = desc1, desc0
            else:  # if name == 'self':
                src0, src1 = desc0, desc1
            delta0, delta1 = layer(desc0, src0), layer(desc1, src1)
            desc0, desc1 = (desc0 + delta0), (desc1 + delta1)
        return desc0, desc1


def log_sinkhorn_iterations(Z: torch.Tensor, log_mu: torch.Tensor, log_nu: torch.Tensor, iters: int) -> torch.Tensor:
    """ Perform Sinkhorn Normalization in Log-space for stability"""
    u, v = torch.zeros_like(log_mu), torch.zeros_like(log_nu)
    for _ in range(iters):
        u = log_mu - torch.logsumexp(Z + v.unsqueeze(1), dim=2)
        v = log_nu - torch.logsumexp(Z + u.unsqueeze(2), dim=1)
    return Z + u.unsqueeze(2) + v.unsqueeze(1)


def log_optimal_transport(scores: torch.Tensor, alpha: torch.Tensor, iters: int) -> torch.Tensor:
    """ Perform Differentiable Optimal Transport in Log-space for stability"""
    b, m, n = scores.shape
    one = scores.new_tensor(1)
    ms, ns = (m*one).to(scores), (n*one).to(scores)

    bins0 = alpha.expand(b, m, 1)
    bins1 = alpha.expand(b, 1, n)
    alpha = alpha.expand(b, 1, 1)

    couplings = torch.cat([torch.cat([scores, bins0], -1),
                           torch.cat([bins1, alpha], -1)], 1)

    norm = - (ms + ns).log()
    log_mu = torch.cat([norm.expand(m), ns.log()[None] + norm])
    log_nu = torch.cat([norm.expand(n), ms.log()[None] + norm])
    log_mu, log_nu = log_mu[None].expand(b, -1), log_nu[None].expand(b, -1)

    Z = log_sinkhorn_iterations(couplings, log_mu, log_nu, iters)
    Z = Z - norm  # multiply probabilities by M+N
    return Z


def arange_like(x, dim: int):
    return x.new_ones(x.shape[dim]).cumsum(0) - 1  # traceable in 1.1


class SuperGlue(nn.Module):
    """SuperGlue feature matching middle-end

    Given two sets of keypoints and locations, we determine the
    correspondences by:
      1. Keypoint Encoding (normalization + visual feature and location fusion)
      2. Graph Neural Network with multiple self and cross-attention layers
      3. Final projection layer
      4. Optimal Transport Layer (a differentiable Hungarian matching algorithm)
      5. Thresholding matrix based on mutual exclusivity and a match_threshold

    The correspondence ids use -1 to indicate non-matching points.

    Paul-Edouard Sarlin, Daniel DeTone, Tomasz Malisiewicz, and Andrew
    Rabinovich. SuperGlue: Learning Feature Matching with Graph Neural
    Networks. In CVPR, 2020. https://arxiv.org/abs/1911.11763

    """
    default_config = {
        'descriptor_dim': 256,
        'weights': 'indoor',
        'keypoint_encoder': [32, 64, 128, 256],
        'GNN_layers': ['self', 'cross'] * 9,
        'sinkhorn_iterations': 100,
        'match_threshold': 0.2,
        'model_path': '',
    }

    def __init__(self, config):
        super().__init__()
        self.config = {**self.default_config, **config}

        self.kenc = KeypointEncoder(
            self.config['descriptor_dim'], self.config['keypoint_encoder'])

        self.gnn = AttentionalGNN(
            feature_dim=self.config['descriptor_dim'], layer_names=self.config['GNN_layers'])

        self.final_proj = nn.Conv1d(
            self.config['descriptor_dim'], self.config['descriptor_dim'],
            kernel_size=1, bias=True)

        bin_score = torch.nn.Parameter(torch.tensor(1.))
        self.register_parameter('bin_score', bin_score)

        self.load_state_dict(torch.load(config['model_path']))

    def forward(self, data):
        """Run SuperGlue on a pair of keypoints and descriptors"""
        desc0, desc1 = data['descriptors0'], data['descriptors1']
        kpts0, kpts1 = data['keypoints0'], data['keypoints1']

        if kpts0.shape[1] == 0 or kpts1.shape[1] == 0:  # no keypoints
            shape0, shape1 = kpts0.shape[:-1], kpts1.shape[:-1]
            return {
                'matches0': kpts0.new_full(shape0, -1, dtype=torch.int),
                'matches1': kpts1.new_full(shape1, -1, dtype=torch.int),
                'matching_scores0': kpts0.new_zeros(shape0),
                'matching_scores1': kpts1.new_zeros(shape1),
            }

        # Keypoint normalization.
        kpts0 = normalize_keypoints(kpts0, data['image0'].shape)
        kpts1 = normalize_keypoints(kpts1, data['image1'].shape)

        # Keypoint MLP encoder.
        desc0 = desc0 + self.kenc(kpts0, data['scores0'])
        desc1 = desc1 + self.kenc(kpts1, data['scores1'])

        # Multi-layer Transformer network.
        desc0, desc1 = self.gnn(desc0, desc1)

        # Final MLP projection.
        mdesc0, mdesc1 = self.final_proj(desc0), self.final_proj(desc1)

        # Compute matching descriptor distance.
        scores = torch.einsum('bdn,bdm->bnm', mdesc0, mdesc1)
        scores = scores / self.config['descriptor_dim']**.5

        # Run the optimal transport.
        scores = log_optimal_transport(
            scores, self.bin_score,
            iters=self.config['sinkhorn_iterations'])

        # Get the matches with score above "match_threshold".
        max0, max1 = scores[:, :-1, :-1].max(2), scores[:, :-1, :-1].max(1)
        indices0, indices1 = max0.indices, max1.indices
        mutual0 = arange_like(indices0, 1)[None] == indices1.gather(1, indices0)
        mutual1 = arange_like(indices1, 1)[None] == indices0.gather(1, indices1)
        zero = scores.new_tensor(0)
        mscores0 = torch.where(mutual0, max0.values.exp(), zero)
        mscores1 = torch.where(mutual1, mscores0.gather(1, indices1), zero)
        valid0 = mutual0 & (mscores0 > self.config['match_threshold'])
        valid1 = mutual1 & valid0.gather(1, indices1)
        indices0 = torch.where(valid0, indices0, indices0.new_tensor(-1))
        indices1 = torch.where(valid1, indices1, indices1.new_tensor(-1))

        return {
            'matches0': indices0, # use -1 for invalid match
            'matches1': indices1, # use -1 for invalid match
            'matching_scores0': mscores0,
            'matching_scores1': mscores1,
        }
